#! /usr/bin/env python
"""Unit tests for landlab.io.netcdf module."""
import numpy as np
import pytest
import xarray as xr
from numpy.testing import assert_array_equal

from landlab import HexModelGrid, RasterModelGrid
from landlab.io.netcdf import from_netcdf, to_netcdf


def test_netcdf_write_int64(tmpdir, format):
    grid = RasterModelGrid((4, 3))
    grid.add_field("topographic__elevation", np.arange(12, dtype=np.int64), at="node")

    with tmpdir.as_cwd():
        to_netcdf(grid, "test.nc", format=format)

        actual = xr.open_dataset("test.nc")
        values = actual["at_node:topographic__elevation"]
        assert_array_equal(values, grid.at_node["topographic__elevation"])
        if format == "NETCDF4":
            assert values.dtype == "int64"
        else:
            assert values.dtype == "int32"


def test_netcdf_write_uint8(tmpdir, format):
    grid = RasterModelGrid((4, 3))
    grid.add_field("topographic__elevation", np.arange(12, dtype=np.uint8), at="node")

    with tmpdir.as_cwd():
        to_netcdf(grid, "test.nc", format=format)

        actual = xr.open_dataset("test.nc")["at_node:topographic__elevation"]
        assert_array_equal(actual, grid.at_node["topographic__elevation"])
        assert actual.dtype == np.uint8 if format == "NETCDF4" else np.int8


@pytest.mark.parametrize("dtype", ("int32", "float32", "float64"))
def test_netcdf_write_dtype(tmpdir, format, dtype):
    """Test write_netcdf with a grid that has an uint8 field."""
    grid = RasterModelGrid((4, 3))
    grid.add_field("topographic__elevation", np.arange(12, dtype=dtype), at="node")

    with tmpdir.as_cwd():
        to_netcdf(grid, "test.nc", format=format)
        actual = xr.open_dataset("test.nc")["at_node:topographic__elevation"]

        assert_array_equal(actual, grid.at_node["topographic__elevation"])
        assert actual.dtype == dtype


def test_at_keyword(tmpdir, at):
    grid = RasterModelGrid((4, 3))

    name = "topographic__elevation"
    for src_at in {"node", "link", "patch", "corner", "face", "cell"}:
        grid.add_field(name, grid.ones(at=src_at) * 10.0, at=src_at)

    include = f"at_{at}:*"
    with tmpdir.as_cwd():
        to_netcdf(grid, "test.nc", format="NETCDF4", include=include)

        with xr.open_dataset("test.nc") as actual:
            actual_fields = {
                name for name in actual.variables if name.startswith("at_")
            }
            nc_name = f"at_{at}:{name}"

            assert actual_fields == {nc_name}
            assert_array_equal(actual[nc_name], getattr(grid, "at_" + at)[name])


def test_raster_model_grid(tmpdir, format):
    grid = RasterModelGrid((4, 3), xy_spacing=(2, 5), xy_of_lower_left=(-2.0, 10.0))
    with tmpdir.as_cwd():
        to_netcdf(grid, "test.nc", format=format)
        actual = from_netcdf("test.nc")
        assert (actual.dx, actual.dy) == (grid.dx, grid.dy)
        assert actual.xy_of_lower_left == grid.xy_of_lower_left


@pytest.mark.parametrize("orientation", ("horizontal", "vertical"))
@pytest.mark.parametrize("node_layout", ("rect", "hex"))
def test_hex_model_grid(tmpdir, format, orientation, node_layout):
    grid = HexModelGrid(
        shape=(4, 5),
        spacing=2.0,
        xy_of_lower_left=(-3, -5),
        orientation=orientation,
        node_layout=node_layout,
    )
    with tmpdir.as_cwd():
        to_netcdf(grid, "test.nc", format=format)
        actual = from_netcdf("test.nc")

        assert actual.spacing == grid.spacing
        assert actual.xy_of_lower_left == grid.xy_of_lower_left
        assert actual.orientation == grid.orientation
        assert actual.node_layout == grid.node_layout


def test_layers(tmpdir, format):
    grid = RasterModelGrid((3, 4))
    grid.event_layers.add(10.0, water_depth=[1.0, 2.0])
    with tmpdir.as_cwd():
        to_netcdf(grid, "test.nc", include="at_layer*", format=format)
        actual = xr.open_dataset("test.nc")
        actual_fields = {name for name in actual.variables if name.startswith("at_")}
        assert actual_fields == {"at_layer_cell:water_depth", "at_layer_cell:thickness"}


def test_layers_append(tmpdir, format):
    grid = RasterModelGrid((3, 4))
    grid.event_layers.add(10.0, water_depth=[1.0, 2.0])
    with tmpdir.as_cwd():
        to_netcdf(grid, "test.nc", include="at_layer*", format=format)
        to_netcdf(grid, "test.nc", include="at_layer*", format=format, mode="a")

        actual = xr.open_dataset("test.nc")
        actual_fields = {name for name in actual.variables if name.startswith("at_")}
        assert actual_fields == {"at_layer_cell:water_depth", "at_layer_cell:thickness"}


@pytest.mark.parametrize("mode", ("w", "a"))
def test_with_and_without_time(tmpdir, format, mode):
    grid = RasterModelGrid((3, 4))
    grid.add_full("elevation", 1.0, at="node")
    with tmpdir.as_cwd():
        to_netcdf(grid, "test-without-time.nc", format=format, mode=mode)
        with xr.open_dataset("test-without-time.nc") as actual:
            assert "time" not in actual.dims
            assert "time" not in actual.variables
            assert actual["at_node:elevation"].dims == ("node",)

        to_netcdf(grid, "test-with-time.nc", format=format, time=10.0, mode=mode)
        with xr.open_dataset("test-with-time.nc") as actual:
            assert "time" in actual.dims
            assert "time" in actual.variables
            assert actual["time"] == [10.0]
            assert actual["at_node:elevation"].dims == ("time", "node")


@pytest.mark.parametrize("mode", ("w", "a"))
@pytest.mark.parametrize("time0", [None, 10.0])
@pytest.mark.parametrize("time1", [None, 100.0])
def test_append_with_and_without_time(tmpdir, format, mode, time0, time1):
    grid = RasterModelGrid((3, 4))
    grid.add_full("elevation", 1.0, at="node")
    with tmpdir.as_cwd():
        to_netcdf(grid, "test.nc", format=format, mode=mode, time=time0)
        to_netcdf(grid, "test.nc", format=format, mode="a", time=time1)

        time0 = np.nan if time0 is None else time0

        with xr.open_dataset("test.nc") as actual:
            assert "time" in actual.dims
            assert "time" in actual.variables
            assert_array_equal(
                actual["time"],
                [
                    np.nan if time0 is None else time0,
                    time0 + 1 if time1 is None else time1,
                ],
            )


def test_append_with_new_field(tmpdir, format):
    grid = RasterModelGrid((3, 4))
    grid.add_full("elevation", 1.0, at="node")
    with tmpdir.as_cwd():
        to_netcdf(grid, "test.nc", format=format, time=0.0)
        grid.add_full("temperature", 2.0, at="node")
        to_netcdf(grid, "test.nc", format=format, mode="a", time=10.0)

        with xr.open_dataset("test.nc") as ds:
            assert sorted(ds.variables) == [
                "at_node:elevation",
                "at_node:temperature",
                "shape",
                "status_at_node",
                "time",
                "xy_of_lower_left",
                "xy_spacing",
            ]
            assert_array_equal(
                ds["at_node:temperature"],
                np.vstack(
                    [np.full(grid.number_of_nodes, np.nan), grid.at_node["temperature"]]
                ),
            )
            assert_array_equal(
                ds["at_node:elevation"],
                np.vstack([grid.at_node["elevation"], grid.at_node["elevation"]]),
            )
            assert_array_equal(ds["time"], [0.0, 10.0])



#! /usr/bin/env python

import os
import pathlib

import numpy
from Cython.Build import cythonize
from setuptools import Extension, setup


def find_extensions(path="."):
    extensions = pathlib.Path(path).rglob("*.pyx")
    return [
        Extension(str(ext.with_suffix("")).replace(os.path.sep, "."), [str(ext)])
        for ext in extensions
    ]


setup(
    include_dirs=[numpy.get_include()],
    ext_modules=cythonize(
        find_extensions("landlab") + find_extensions("tests"),
        compiler_directives={"embedsignature": True, "language_level": 3},
    ),
)



"""
Unit tests for landlab.components.dimensionless.discharge
"""

import numpy as np
from numpy.testing import assert_array_almost_equal

from landlab import RasterModelGrid
from landlab.components.dimensionless_discharge import DimensionlessDischarge


def init_grid():
    """initialize grid for testing"""
    watershed_grid = RasterModelGrid((3, 3))
    watershed_grid.add_ones("surface_water__unit_discharge", at="node")
    watershed_grid.add_ones("channel_bottom_sediment_grain__d50_diameter", at="node")
    watershed_grid.add_ones("topographic__elevation", at="node")
    watershed_grid.at_node["topographic__elevation"] = np.array(
        [[1.1, 2, 3, 4, 2, 3, 4, 5, 3]]
    )
    return watershed_grid


def test_dimensionless_discharge_final_values():
    """Testing for correct final dimensionless discharge values"""
    watershed_grid = init_grid()
    # Setting the gravitation constant to 9.8 to make hand calculating
    # expeted values easier.
    dd = DimensionlessDischarge(watershed_grid, gravity=9.8)
    dd.run_one_step()
    # 0.55372743 is the expected value from the dimensionless discharge
    # (q*) equation in the Tang et al paper when surface_water__unit_discharge = 1
    # and channel_bottom_sediment_grain__d50_diameter = 1.
    # All values in dimensionless_discharge should be 0.5537274 after
    # running the un_one_step() function since the D50 and
    # surface_water__unit_discharge values are all 1s
    expected_values = [
        0.55372743,
        0.55372743,
        0.55372743,
        0.55372743,
        0.55372743,
        0.55372743,
        0.55372743,
        0.55372743,
        0.55372743,
    ]
    assert_array_almost_equal(
        watershed_grid.at_node["dimensionless_discharge"], expected_values
    )


def test_dimensionless_discharge_threshold_values():
    """Validate threshold values are correctly calculated"""
    watershed_grid = init_grid()
    # Setting the gravitation constant to 9.8 to make hand calculating
    # expeted values easier.
    dd = DimensionlessDischarge(watershed_grid, gravity=9.8)
    dd.run_one_step()
    expected_values = [
        11.09442633,
        12.63600546,
        14.73515969,
        11.01921767,
        11.72461201,
        12.5362998,
        10.94510988,
        10.94510988,
        10.94510988,
    ]
    assert_array_almost_equal(
        watershed_grid.at_node["dimensionless_discharge_threshold"],
        expected_values,
    )



import numpy as np
cimport numpy as np
cimport cython


DTYPE_INT = int
ctypedef np.int_t DTYPE_INT_t

DTYPE_FLOAT = np.float64
ctypedef np.float64_t DTYPE_FLOAT_t


cpdef _thresholder(np.ndarray[DTYPE_INT_t, ndim=1] stack,
             np.ndarray[DTYPE_INT_t, ndim=1] link_to_rcvr,
             np.ndarray[DTYPE_INT_t, ndim=1] receivers,
             np.ndarray[DTYPE_FLOAT_t, ndim=1] linkLengths,
             np.ndarray[DTYPE_FLOAT_t, ndim=1] el,
             DTYPE_FLOAT_t slope_thres):

    """
    Calcualte D8 flow dirs
    stack: the flow upstream node order
    link_to_rcvr: Link to receiver
    receivers: receivers
    linkLengths: length of links
    el: topographic elevation
    """
    cdef int node

    for node in stack:
        dist_to_rcvr = linkLengths[link_to_rcvr[node]]
        el[node]  = np.minimum(el[node],el[receivers[node]] + slope_thres*dist_to_rcvr)



import os
import pathlib
import shutil
import tempfile

import nox

PROJECT = "landlab"
ROOT = pathlib.Path(__file__).parent


@nox.session(venv_backend="mamba")
def test(session: nox.Session) -> None:
    """Run the tests."""
    session.conda_install("--file", "requirements.txt")
    session.conda_install("--file", "requirements-testing.txt")
    session.conda_install("richdem")
    session.install("-e", ".", "--no-deps")

    args = [
        "-n",
        "auto",
        "--cov",
        PROJECT,
        "-vvv",
        # "--dist", "worksteal",  # this is not available quite yet
    ] + session.posargs

    if "CI" in os.environ:
        args.append(f"--cov-report=xml:{ROOT.absolute()!s}/coverage.xml")
    session.run("pytest", *args)

    if "CI" not in os.environ:
        session.run("coverage", "report", "--ignore-errors", "--show-missing")


@nox.session(name="test-notebooks", venv_backend="mamba")
def test_notebooks(session: nox.Session) -> None:
    """Run the notebooks."""
    args = [
        "pytest",
        "notebooks",
        "--nbmake",
        "--nbmake-kernel=python3",
        "--nbmake-timeout=3000",
        "-n",
        "auto",
        "-vvv",
    ] + session.posargs

    session.install("git+https://github.com/mcflugen/nbmake.git@mcflugen/add-markers")
    session.conda_install("richdem")
    session.conda_install(
        "pytest",
        "pytest-xdist",
        "--file",
        "requirements-notebooks.txt",
        "--file",
        "requirements.txt",
    )
    session.install("-e", ".", "--no-deps")

    session.run(*args)


@nox.session(name="test-cli")
def test_cli(session: nox.Session) -> None:
    """Test the command line interface."""
    session.install(".")
    session.run("landlab", "--help")
    session.run("landlab", "--version")
    session.run("landlab", "index", "--help")
    session.run("landlab", "list", "--help")
    session.run("landlab", "provided-by", "--help")
    session.run("landlab", "provides", "--help")
    session.run("landlab", "used-by", "--help")
    session.run("landlab", "uses", "--help")
    session.run("landlab", "validate", "--help")


@nox.session
def lint(session: nox.Session) -> None:
    """Look for lint."""
    session.install("pre-commit")
    session.run("pre-commit", "run", "--all-files")

    # towncrier(session)


@nox.session
def towncrier(session: nox.Session) -> None:
    """Check that there is a news fragment."""
    session.install("towncrier")
    session.run("towncrier", "check", "--compare-with", "origin/master")


@nox.session(name="build-index")
def build_index(session: nox.Session) -> None:
    session.install(".[docs]")

    with open(ROOT / "docs" / "index.toml", "w") as fp:
        print("# This file was automatically generated with:", file=fp, flush=True)
        print("#     nox -s build-index", file=fp, flush=True)
        session.run(
            "landlab", "--silent", "index", "grids", "fields", "components", stdout=fp
        )


@nox.session(name="build-docs", venv_backend="mamba")
def build_docs(session: nox.Session) -> None:
    """Build the docs."""
    session.conda_install("richdem")
    session.install(".[docs]")

    clean_docs(session)
    session.run(
        "sphinx-build",
        "-b",
        "html",
        "-W",
        "--keep-going",
        str(ROOT / "docs/source"),
        str(ROOT / "build/html"),
    )


@nox.session(name="build-requirements")
def build_requirements(session: nox.Session) -> None:
    """Create requirements files from pyproject.toml."""
    session.install("tomli")

    with open("requirements.txt", "w") as fp:
        session.run("python", "requirements.py", stdout=fp)

    for extra in ["dev", "docs", "notebooks", "testing"]:
        with open(f"requirements-{extra}.txt", "w") as fp:
            session.run("python", "requirements.py", extra, stdout=fp)


@nox.session
def build(session: nox.Session) -> None:
    """Build sdist and wheel dists."""
    session.install("pip")
    session.install("build")
    session.run("python", "--version")
    session.run("pip", "--version")
    session.run("python", "-m", "build", "--outdir", "./build/wheelhouse")


@nox.session
def release(session):
    """Tag, build and publish a new release to PyPI."""
    session.install("zest.releaser[recommended]")
    session.install("zestreleaser.towncrier")
    session.run("fullrelease")


@nox.session(name="publish-testpypi")
def publish_testpypi(session):
    """Publish wheelhouse/* to TestPyPI."""
    session.run("twine", "check", "build/wheelhouse/*")
    session.run(
        "twine",
        "upload",
        "--skip-existing",
        "--repository-url",
        "https://test.pypi.org/legacy/",
        "build/wheelhouse/*.tar.gz",
    )


@nox.session(name="publish-pypi")
def publish_pypi(session):
    """Publish wheelhouse/* to PyPI."""
    session.run("twine", "check", "build/wheelhouse/*")
    session.run(
        "twine",
        "upload",
        "--skip-existing",
        "build/wheelhouse/*.tar.gz",
    )


@nox.session(python=False)
def clean(session):
    """Remove all .venv's, build files and caches in the directory."""
    for folder in _args_to_folders(session.posargs):
        with session.chdir(folder):
            shutil.rmtree("build", ignore_errors=True)
            shutil.rmtree("build/wheelhouse", ignore_errors=True)
            shutil.rmtree(f"{PROJECT}.egg-info", ignore_errors=True)
            shutil.rmtree(".pytest_cache", ignore_errors=True)
            shutil.rmtree(".venv", ignore_errors=True)

            for pattern in ["*.py[co]", "__pycache__"]:
                _clean_rglob(pattern)


@nox.session(python=False, name="clean-checkpoints")
def clean_checkpoints(session):
    """Remove jupyter notebook checkpoint files."""
    for folder in _args_to_folders(session.posargs):
        with session.chdir(folder):
            _clean_rglob("*-checkpoint.ipynb")
            _clean_rglob(".ipynb_checkpoints")


@nox.session(python=False, name="clean-docs")
def clean_docs(session: nox.Session) -> None:
    """Clean up the docs folder."""
    session.chdir(ROOT / "build")
    if os.path.exists("html"):
        shutil.rmtree("html")


@nox.session(python=False, name="clean-ext")
def clean_ext(session: nox.Session) -> None:
    """Clean shared libraries for extension modules."""
    for folder in _args_to_folders(session.posargs):
        with session.chdir(folder):
            _clean_rglob("*.so")


@nox.session(python=False)
def nuke(session):
    """Run all clean sessions."""
    clean_checkpoints(session)
    clean_docs(session)
    clean(session)
    clean_ext(session)


@nox.session(name="list-wheels")
def list_wheels(session):
    print(os.linesep.join(_get_wheels(session)))


@nox.session(name="list-ci-matrix")
def list_ci_matrix(session):
    def _os_from_wheel(name):
        if "linux" in name:
            return "linux"
        elif "macos" in name:
            return "macos"
        elif "win" in name:
            return "windows"

    for wheel in _get_wheels(session):
        print(f"- cibw-only: {wheel}")
        print(f"  os: {_os_from_wheel(wheel)}")


def _get_wheels(session):
    platforms = session.posargs or ["linux", "macos", "windows"]
    session.install("cibuildwheel")

    wheels = []
    for platform in platforms:
        with tempfile.TemporaryFile("w+") as fp:
            session.run(
                "cibuildwheel",
                "--print-build-identifiers",
                "--platform",
                platform,
                stdout=fp,
            )
            fp.seek(0)
            wheels += fp.read().splitlines()
    return wheels


def _args_to_folders(args):
    return [ROOT] if not args else [pathlib.Path(f) for f in args]


def _clean_rglob(pattern):
    nox_dir = pathlib.Path(".nox")

    for p in pathlib.Path(".").rglob(pattern):
        if nox_dir in p.parents:
            continue
        if p.is_dir():
            p.rmdir()
        else:
            p.unlink()



"""Landlab component for overland flow using the linearized diffusion-wave approximation.

Created on Fri May 27 14:26:13 2016

@author: gtucker
"""


import numpy as np

from landlab import Component

_FOUR_THIRDS = 4.0 / 3.0
_SEVEN_THIRDS = 7.0 / 3.0
_MICRO_DEPTH = 1.0e-6  # tiny water depth to avoid blowup in time-step estimator


class LinearDiffusionOverlandFlowRouter(Component):
    r"""Calculate water flow over topography.

    Landlab component that implements a two-dimensional, linearized
    diffusion-wave model. The diffusion-wave approximation is a simplification
    of the shallow-water equations that omits the momentum terms. The flow
    velocity is calculated using the local water-surface slope as an
    approximation of the energy slope. With this linearized form, flow velocity
    is calculated using a linearized Manning equation, with the water-surface
    slope being used as the slope factor. There are two governing equations, one
    that represents conservation of water mass:

    ..math::

        \frac{\partial H}{\partial t} = (P - I) - \nabla\cdot\mathbf{q}

    where :math:`H(x,y,t)` is local water depth, :math:`t` is time, :math:`P`
    is precipitation rate, :math:`I` is infiltration rate, and :math:`\mathbf{q}`
    is specific water discharge, which equals depth times depth-averaged
    velocity. The other governing equation represents specific discharge as a
    function of gravity, pressure, and friction:

    ..math::

        \mathbf{q} = \frac{H^{4/3}}{n^2 U_c} \nabla w

    where :math:`n` is the friction factor ("Manning's n"), :math:`U_c` is a
    characteristic flow velocity, and :math:`w` is water-surface height, which
    is the sum of topographic elevation plus water depth.

    Infiltration rate should decline smoothly to zero as surface water depth
    approaches zero. To ensure this, infiltration rate is calculated as

    ..math::

        I = I_c \left( 1 - e^{-H / H_i} ) \right)

    where :math:`H_i` is a characteristic water depth. The concept here is that
    when :math:`H \le H_i`, small spatial variations in water depth will leave
    parts of the ground un-ponded and therefore not subject to any infiltration.
    Mathematically, :math:`I \approx 0.95 I_c` when :math:`H = 3H_i`.

    Examples
    --------
    >>> from landlab import RasterModelGrid
    >>> grid = RasterModelGrid((3, 3))
    >>> _ = grid.add_zeros('topographic__elevation', at='node')
    >>> olf = LinearDiffusionOverlandFlowRouter(grid, roughness=0.1)
    >>> round(olf.vel_coef)
    100
    >>> olf.rain_rate
    1e-05
    >>> olf.rain_rate = 1.0e-4
    >>> olf.run_one_step(dt=10.0)
    >>> grid.at_node['surface_water__depth'][4]
    0.001

    References
    ----------
    **Required Software Citation(s) Specific to this Component**

    None Listed

    **Additional References**

    None Listed

    """

    _name = "LinearDiffusionOverlandFlowRouter"

    _unit_agnostic = True

    _info = {
        "surface_water__depth": {
            "dtype": float,
            "intent": "out",
            "optional": False,
            "units": "m",
            "mapping": "node",
            "doc": "Depth of water on the surface",
        },
        "surface_water__depth_at_link": {
            "dtype": float,
            "intent": "out",
            "optional": False,
            "units": "m",
            "mapping": "link",
            "doc": "Depth of water on the surface at grid links",
        },
        "topographic__elevation": {
            "dtype": float,
            "intent": "in",
            "optional": False,
            "units": "m",
            "mapping": "node",
            "doc": "Land surface topographic elevation",
        },
        "water__specific_discharge": {
            "dtype": float,
            "intent": "out",
            "optional": False,
            "units": "m2/s",
            "mapping": "link",
            "doc": "flow discharge component in the direction of the link",
        },
        "water__velocity": {
            "dtype": float,
            "intent": "out",
            "optional": False,
            "units": "m/s",
            "mapping": "link",
            "doc": "flow velocity component in the direction of the link",
        },
        "water_surface__gradient": {
            "dtype": float,
            "intent": "out",
            "optional": False,
            "units": "m/s",
            "mapping": "link",
            "doc": "Downstream gradient of the water surface.",
        },
        "water_surface__elevation": {
            "dtype": float,
            "intent": "out",
            "optional": False,
            "units": "m",
            "mapping": "node",
            "doc": "Elevation of the water surface.",
        },
    }

    def __init__(
        self,
        grid,
        roughness=0.01,
        rain_rate=1.0e-5,
        infilt_rate=0.0,
        infilt_depth_scale=0.001,
        velocity_scale=1.0,
        cfl_factor=0.2,
    ):
        """Initialize the LinearDiffusionOverlandFlowRouter.

        Parameters
        ----------
        grid : ModelGrid
            Landlab ModelGrid object
        roughness : float, defaults to 0.01
            Manning roughness coefficient, s/m^1/3
        rain_rate : float, optional (defaults to 36 mm/hr)
            Rainfall rate, m/s
        infilt_depth_scale : float, defaults to 0.001
            Depth scale for water infiltration, m
        infilt_rate : float, optional (defaults to 0)
            Maximum rate of infiltration, m/s
        velocity_scale : float, defaults to 1
            Characteristic flow velocity, m/s
        cfl_factor : float, optional (defaults to 0.2)
            Time-step control factor: fraction of maximum estimated time-step
            that is actually used (must be <=1)
        """
        super().__init__(grid)

        if roughness <= 0.0:
            raise ValueError(f"roughness must be greater than zero {roughness}")
        if rain_rate < 0.0:
            raise ValueError(f"rain_rate must not be negative {rain_rate}")
        if infilt_depth_scale <= 0.0:
            raise ValueError(
                f"infilt_depth_scale must be greater than zero {infilt_depth_scale}"
            )
        if infilt_rate < 0.0:
            raise ValueError(f"infilt_rate must not be negative {infilt_rate}")
        if velocity_scale <= 0.0:
            raise ValueError(
                f"velocity_scale must be greater than zero {velocity_scale}"
            )
        if cfl_factor > 1.0 or cfl_factor <= 0.0:
            raise ValueError(f"cfl_factor must >0, <=1 {cfl_factor}")

        # Store parameters and do unit conversion
        self._rain = rain_rate
        self._infilt = infilt_rate
        self._infilt_depth_scale = infilt_depth_scale
        self._vel_coef = 1.0 / (roughness**2 * velocity_scale)

        self._elev = grid.at_node["topographic__elevation"]

        self.initialize_output_fields()
        self._depth = grid.at_node["surface_water__depth"]
        self._depth_at_link = grid.at_link["surface_water__depth_at_link"]
        self._vel = grid.at_link["water__velocity"]
        self._disch = grid.at_link["water__specific_discharge"]
        self._wsgrad = grid.at_link["water_surface__gradient"]
        self._water_surf_elev = grid.at_node["water_surface__elevation"]

        self._inactive_links = grid.status_at_link == grid.BC_LINK_IS_INACTIVE

        self._cfl_param = cfl_factor * 0.5 * np.amin(grid.length_of_link) ** 2

    @property
    def rain_rate(self):
        """Rainfall rate"""
        return self._rain

    @rain_rate.setter
    def rain_rate(self, value):
        if value < 0.0:
            raise ValueError(f"rain_rate must be positive {value}")
        self._rain = value

    @property
    def vel_coef(self):
        """Velocity coefficient.

        (1/(roughness^2 x velocity_scale)
        """
        return self._vel_coef

    def _cfl_time_step(self):
        """Calculate maximum time-step size using CFL criterion for explicit
        FTCS diffusion."""
        max_water_depth = np.amax(self._depth, initial=_MICRO_DEPTH)
        max_diffusivity = self._vel_coef * max_water_depth**_SEVEN_THIRDS
        return self._cfl_param / max_diffusivity

    def update_for_one_iteration(self, iter_dt):
        """Update state variables for one iteration of duration iter_dt."""

        # Calculate the water-surface elevation
        self._water_surf_elev[:] = self._elev + self._depth

        # Calculate water depth at links. This implements an "upwind" scheme
        # in which water depth at the links is the depth at the higher of the
        # two nodes.
        self._grid.map_value_at_max_node_to_link(
            self._water_surf_elev, "surface_water__depth", out=self._depth_at_link
        )

        # Calculate the water-surface gradient and impose any closed boundaries
        self.grid.calc_grad_at_link(self._water_surf_elev, out=self._wsgrad)
        self._wsgrad[self._inactive_links] = 0.0

        # Calculate velocity using the linearized Manning equation.
        self._vel[:] = (
            -self._vel_coef * self._depth_at_link**_FOUR_THIRDS * self._wsgrad
        )

        # Calculate discharge
        self._disch[:] = self._depth_at_link * self._vel

        # Flux divergence
        dqda = self._grid.calc_flux_div_at_node(self._disch)

        # Rates of infiltration and runoff
        infilt = self._infilt * (1.0 - np.exp(-self._depth / self._infilt_depth_scale))

        # Rate of change of water depth
        dHdt = self._rain - infilt - dqda

        # Update water depth: simple forward Euler scheme
        self._depth[self._grid.core_nodes] += dHdt[self._grid.core_nodes] * iter_dt

        # Very crude numerical hack: prevent negative water depth (TODO: better)
        self._depth.clip(min=0.0, out=self._depth)

    def run_one_step(self, dt):
        """Calculate water flow for a time period `dt`.

        Default units for dt are *seconds*. We use a time-step subdivision
        algorithm that ensures step size is always below CFL limit.
        """
        remaining_time = dt
        while remaining_time > 0.0:
            dtmax = self._cfl_time_step()  # biggest possible time step size
            dt_this_iter = min(dtmax, remaining_time)  # step we'll actually use
            self.update_for_one_iteration(dt_this_iter)
            remaining_time -= dt_this_iter



#! /usr/bin/env python
import numpy as np
import pytest
from numpy.testing import assert_array_equal

from landlab.field.graph_field import FieldDataset


def test_init():
    ds = FieldDataset("node")
    assert ds.size is None
    assert ds.fixed_size is True
    assert list(ds) == []
    assert len(ds) == 0


def test_init_with_size():
    ds = FieldDataset("node", size=1)
    assert ds.size == 1
    with pytest.raises(ValueError):
        FieldDataset("node", size=1.1)
    with pytest.raises(ValueError):
        FieldDataset("node", size=-1)


def test_init_fixed_size():
    assert FieldDataset("node", fixed_size=True).fixed_size is True
    assert FieldDataset("node", fixed_size=1).fixed_size is True
    assert FieldDataset("node", fixed_size=False).fixed_size is False
    assert FieldDataset("node", fixed_size=0).fixed_size is False
    assert FieldDataset("node", fixed_size="true").fixed_size is True
    assert FieldDataset("node", fixed_size="false").fixed_size is True


def test_size_setter():
    ds = FieldDataset("node", fixed_size=True)
    assert ds.size is None

    ds.size = 3
    assert ds.size == 3
    with pytest.raises(ValueError):
        ds.size = 4
    ds.size = 3

    ds.fixed_size = False
    ds.size = 4
    assert ds.size == 4

    with pytest.raises(ValueError):
        ds.size = 1.1
    with pytest.raises(ValueError):
        ds.size = -1


def test_add_field():
    ds = FieldDataset("node")
    assert ds.size is None
    ds.set_value("air_temperature", [0, 0, 0])
    assert ds.size == 3
    assert_array_equal(ds["air_temperature"], (0, 0, 0))
    assert len(ds) == 1
    assert list(ds) == ["air_temperature"]


def test_field_maintains_reference():
    ds = FieldDataset("node")
    array = np.arange(3)
    ds.set_value("air_temperature", array)
    assert ds["air_temperature"] is array


def test_setitem():
    ds = FieldDataset("node")
    assert ds.size is None
    ds["air_temperature"] = [0, 0, 0]
    assert ds.size == 3
    assert_array_equal(ds["air_temperature"], (0, 0, 0))
    assert len(ds) == 1
    assert list(ds) == ["air_temperature"]


def test_add_fields():
    ds = FieldDataset("node", fixed_size=True)
    ds.set_value("air_temperature", [0, 0, 0])
    ds.set_value("ground_temperature", [1, 1, 1])
    assert_array_equal(ds["air_temperature"], (0, 0, 0))
    assert_array_equal(ds["ground_temperature"], (1, 1, 1))
    assert len(ds) == 2
    assert sorted(ds) == ["air_temperature", "ground_temperature"]


def test_add_fields_fixed_size():
    ds = FieldDataset("node", fixed_size=True)
    ds.set_value("air_temperature", [0, 0, 0])
    with pytest.raises(ValueError):
        ds.set_value("ground_temperature", [1, 1, 1, 1])


def test_add_fields_adjustable_size():
    ds = FieldDataset("node", fixed_size=False)
    ds.set_value("air_temperature", [0, 0, 0])
    assert ds.size == 3
    ds.set_value("air_temperature", [1, 1, 1, 1])
    assert ds.size == 4
    with pytest.raises(ValueError):
        ds.set_value("ground_temperature", [1, 1, 1])


def test_add_fields_fixed_size_setter():
    ds = FieldDataset("node", fixed_size=False)
    ds.set_value("air_temperature", [0, 0, 0])
    ds.set_value("air_temperature", [1, 1, 1, 1])
    ds.fixed_size = True
    with pytest.raises(ValueError):
        ds.set_value("air_temperature", [1, 1, 1])
    assert ds.size == 4
    ds.fixed_size = False
    ds.set_value("air_temperature", [1, 1, 1])
    assert ds.size == 3


def test_pop():
    ds = FieldDataset("node")
    ds.set_value("air_temperature", [0, 0, 0])
    ds.set_value("ground_temperature", [1, 1, 1])
    assert sorted(ds) == ["air_temperature", "ground_temperature"]

    assert_array_equal(ds.pop("ground_temperature"), [1, 1, 1])
    assert sorted(ds) == ["air_temperature"]

    assert_array_equal(ds.pop("air_temperature"), [0, 0, 0])
    assert sorted(ds) == []

    with pytest.raises(KeyError):
        ds.pop("air_temperature")


def test_contains():
    ds = FieldDataset("node")
    ds.set_value("air_temperature", [0, 0, 0])
    assert "air_temperature" in ds
    assert "ground_temperature" not in ds
    assert "node" not in ds


def test_scalars():
    ds = FieldDataset("node")
    ds["air_temperature"] = 2
    assert ds["air_temperature"] == 2
    assert ds.size == 1

    ds["ground_temperature"] = [0]
    assert ds["ground_temperature"] == [0]


def test_size_is_one():
    ds = FieldDataset("grid")
    ds.size = 1
    ds["air_temperature"] = [1, 2, 3]
    assert_array_equal(ds["air_temperature"], [[1, 2, 3]])
    ds["ground_temperature"] = [0, 0, 0, 0, 0]
    assert_array_equal(ds["ground_temperature"], [[0, 0, 0, 0, 0]])
    ds["gravity"] = 9.81
    assert ds["gravity"] == pytest.approx(9.81)


def test_reshaping_arrays():
    ds = FieldDataset("links", size=3, fixed_size=True)
    ds["air_temperature"] = [1.0, 2.0, 3.0]
    assert_array_equal(ds["air_temperature"], [1.0, 2.0, 3.0])

    assert ds.size == 3
    ds["nodes"] = [[0, 0], [0, 0], [0, 0]]
    assert_array_equal(ds["nodes"], [[0, 0], [0, 0], [0, 0]])

    ds["nodes"] = [1, 2, 3, 4, 5, 6]
    assert_array_equal(ds["nodes"], [[1, 2], [3, 4], [5, 6]])

    with pytest.raises(ValueError):
        ds["nodes"] = [1, 2, 3, 4, 5, 6, 7]



import pytest

from landlab import RasterModelGrid
from landlab.components import FlowAccumulator, HackCalculator


@pytest.fixture()
def simple_hack_test_grid():
    mg = RasterModelGrid((10, 10))
    z = mg.add_zeros("topographic__elevation", at="node")
    z += mg.x_of_node + mg.y_of_node
    fa = FlowAccumulator(mg)
    fa.run_one_step()
    return mg


def test_no_full_hack_dataframe_specified(simple_hack_test_grid):
    mg = simple_hack_test_grid
    hc = HackCalculator(mg)
    with pytest.raises(NotImplementedError):
        hc.full_hack_dataframe


def test_no_full_hack_dataframe_built(simple_hack_test_grid):
    mg = simple_hack_test_grid
    hc = HackCalculator(mg, save_full_df=True)
    with pytest.raises(RuntimeError):
        hc.full_hack_dataframe


def test_no_coefficient_df_built(simple_hack_test_grid):
    mg = simple_hack_test_grid
    hc = HackCalculator(mg)
    with pytest.raises(RuntimeError):
        hc.hack_coefficient_dataframe



#! /usr/bin/env python
r"""Registry of landlab components being used.

The landlab registry keeps track of landlab components that have
be instantiated by a user. A user can then get a list of all
the components in use and then print a list of citations for
all of the components they have used.

Examples
--------
>>> from landlab import registry

>>> registry.registered
('landlab',)
>>> print(registry.format_citations())
# Citations
## landlab
    @article{hobley2017creative,
    title={Creative computing with Landlab: an open-source toolkit
        for building, coupling, and exploring two-dimensional
        numerical models of Earth-surface dynamics},
    author={Hobley, Daniel EJ and Adams, Jordan M and Nudurupati,
        Sai Siddhartha and Hutton, Eric WH and Gasparini, Nicole M and
        Istanbulluoglu, Erkan and Tucker, Gregory E},
    journal={Earth Surface Dynamics},
    volume={5},
    number={1},
    pages={21},
    year={2017},
    publisher={Copernicus GmbH}
    }
    @article{barnhart2020short,
    author = {Barnhart, K. R. and Hutton, E. W. H. and Tucker, G. E.
        and Gasparini, N. M. and Istanbulluoglu, E. and Hobley,
        D. E. J. and Lyons, N. J. and Mouchene, M. and Nudurupati,
        S. S. and Adams, J. M. and Bandaragoda, C.},
    title = {Short communication: Landlab v2.0: A software package
        for Earth surface dynamics},
    journal = {Earth Surface Dynamics Discussions},
    volume = {2020},
    year = {2020},
    pages = {1--25},
    url = {https://www.earth-surf-dynam-discuss.net/esurf-2020-12/},
    doi = {10.5194/esurf-2020-12}
    }

When a component contains citation information, and the component has been
instantiated (not just imported) the component citation is also included.

>>> from landlab import RasterModelGrid
>>> from landlab.components import Flexure

>>> grid = RasterModelGrid((4, 5))
>>> _ = grid.add_zeros("lithosphere__overlying_pressure_increment", at="node")
>>> _ = grid.add_zeros("lithosphere_surface__elevation_increment", at="node")
>>> flexure = Flexure(grid)
>>> print(registry.format_citations())
# Citations
## landlab
    @article{hobley2017creative,
    title={Creative computing with Landlab: an open-source toolkit
        for building, coupling, and exploring two-dimensional
        numerical models of Earth-surface dynamics},
    author={Hobley, Daniel EJ and Adams, Jordan M and Nudurupati,
        Sai Siddhartha and Hutton, Eric WH and Gasparini, Nicole M and
        Istanbulluoglu, Erkan and Tucker, Gregory E},
    journal={Earth Surface Dynamics},
    volume={5},
    number={1},
    pages={21},
    year={2017},
    publisher={Copernicus GmbH}
    }
    @article{barnhart2020short,
    author = {Barnhart, K. R. and Hutton, E. W. H. and Tucker, G. E.
        and Gasparini, N. M. and Istanbulluoglu, E. and Hobley,
        D. E. J. and Lyons, N. J. and Mouchene, M. and Nudurupati,
        S. S. and Adams, J. M. and Bandaragoda, C.},
    title = {Short communication: Landlab v2.0: A software package
        for Earth surface dynamics},
    journal = {Earth Surface Dynamics Discussions},
    volume = {2020},
    year = {2020},
    pages = {1--25},
    url = {https://www.earth-surf-dynam-discuss.net/esurf-2020-12/},
    doi = {10.5194/esurf-2020-12}
    }
<BLANKLINE>
## Flexure
    @article{hutton2008sedflux,
    title={Sedflux 2.0: An advanced process-response model that
        generates three-dimensional stratigraphy},
    author={Hutton, Eric WH and Syvitski, James PM},
    journal={Computers \& Geosciences},
    volume={34},
    number={10},
    pages={1319--1337},
    year={2008},
    publisher={Pergamon}
    }

Finally, the component's citation information is accessible through an
attribute called ``cite_as``:

>>> print(Flexure.cite_as)
    @article{hutton2008sedflux,
    title={Sedflux 2.0: An advanced process-response model that
        generates three-dimensional stratigraphy},
    author={Hutton, Eric WH and Syvitski, James PM},
    journal={Computers \& Geosciences},
    volume={34},
    number={10},
    pages={1319--1337},
    year={2008},
    publisher={Pergamon}
    }

"""

import os

from . import _info
from .core.messages import indent_and_wrap


class ComponentRegistry:

    """A registry for instantiated landlab components."""

    def __init__(self, objs=None):
        self._registered = []
        if objs is not None:
            try:
                [self.add(obj) for obj in objs]
            except TypeError:
                self.add(objs)

    def add(self, cls):
        """Add a class to the registry.

        Parameters
        ----------
        cls : Component
            A landlab component to register as used.
        """
        if cls not in self._registered:
            self._registered.append(cls)

    @property
    def registered(self):
        """All registered classes.

        Returns
        -------
        tuple
            The components in the registry.

        Examples
        --------
        >>> from landlab._registry import ComponentRegistry
        >>> registry = ComponentRegistry()
        >>> registry.registered
        ()
        >>> class FooBar(object):
        ...    pass
        >>> registry.add(FooBar)
        >>> registry.registered
        ('FooBar',)
        """
        return tuple(ComponentRegistry.get_name(obj) for obj in self._registered)

    @staticmethod
    def format_citation(obj):
        """Format a single citation.

        Parameters
        ----------
        obj : Component
            A landlab component class or instance.

        Returns
        -------
        str
            The formatted citation, or "None" if there is no citation
            given.

        Examples
        --------
        >>> from landlab._registry import ComponentRegistry
        >>> registry = ComponentRegistry()
        >>> class DoNothingComponent(object):
        ...     pass
        >>> print(registry.format_citation(DoNothingComponent))
        ## DoNothingComponent
            None

        >>> class SorterAndSearcher(object):
        ...     _cite_as = '''
        ... @book{knuth1998art,
        ... title={The art of computer programming: sorting and searching},
        ... author={Knuth, Donald Ervin},
        ... volume={3},
        ... year={1998},
        ... publisher={Pearson Education}
        ... }'''
        >>> print(registry.format_citation(SorterAndSearcher))
        ## SorterAndSearcher
            @book{knuth1998art,
            title={The art of computer programming: sorting and searching},
            author={Knuth, Donald Ervin},
            volume={3},
            year={1998},
            publisher={Pearson Education}
            }
        """
        name = ComponentRegistry.get_name(obj)
        header = [f"## {name}"]

        cite_as = ComponentRegistry.get_citations(obj)

        body = []
        for citation in cite_as:
            body.append(indent_and_wrap(citation, indent=" " * 4))

        return os.linesep.join(header + body)

    @staticmethod
    def get_name(obj):
        """Get the display name for an object.

        Examples
        --------
        >>> from landlab._registry import ComponentRegistry
        >>> class MontyPython(object):
        ...     name = "Eric Idle"
        >>> ComponentRegistry.get_name(MontyPython)
        'Eric Idle'
        >>> class MontyPython(object):
        ...     _name = "Graham Chapman"
        >>> ComponentRegistry.get_name(MontyPython)
        'Graham Chapman'
        >>> class MontyPython(object):
        ...     pass
        >>> ComponentRegistry.get_name(MontyPython)
        'MontyPython'

        """
        name = "Unknown"
        for attr in ("name", "_name", "__name__"):
            try:
                name = getattr(obj, attr)
            except AttributeError:
                pass
            else:
                break
        return name

    @staticmethod
    def get_citations(obj):
        """Get a list of citations from an object."""
        citations = "None"
        for attr in ("cite_as", "_cite_as"):
            try:
                citations = getattr(obj, attr)
            except AttributeError:
                pass
            else:
                break
        if isinstance(citations, str):
            citations = [citations]
        return citations

    def format_citations(self):
        """Format citations for all registered components.

        Returns
        -------
        str
            The formatted citations.

        Examples
        --------
        >>> from landlab._registry import ComponentRegistry
        >>> registry = ComponentRegistry()

        >>> class HolyGrailFinder(object):
        ...     _name = 'Monty Python'
        ...     _cite_as = ['''@book{python2000holy,
        ...         title={the Holy Grail},
        ...         author={Python, Monty and Chapman, Graham and Cleese, John and Gilliam, Terry and Jones, Terry and Idle, Eric and Palin, Michael},
        ...         year={2000},
        ...         publisher={EMI Records}
        ...         }''',
        ...         '''@book{chapman1989complete,
        ...         title={The Complete Monty Python's Flying Circus: All the Words. Volume one},
        ...         author={Chapman, Graham and Python, Monty},
        ...         volume={1},
        ...         year={1989},
        ...         publisher={Pantheon}
        ...         }''']
        >>> class Evolution(object):
        ...     _cite_as = '''
        ...         @book{darwin1859origin,
        ...         title={On the origin of species},
        ...         author={Darwin, Charles},
        ...         year={1859},
        ...         publisher={Lulu. com}
        ...         }'''
        >>> registry.add(HolyGrailFinder)
        >>> registry.add(Evolution)
        >>> print(registry.format_citations())
        # Citations
        ## Monty Python
            @book{python2000holy,
            title={the Holy Grail},
            author={Python, Monty and Chapman, Graham and Cleese, John and
                Gilliam, Terry and Jones, Terry and Idle, Eric and Palin,
                Michael},
            year={2000},
            publisher={EMI Records}
            }
            @book{chapman1989complete,
            title={The Complete Monty Python's Flying Circus: All the Words.
                Volume one},
            author={Chapman, Graham and Python, Monty},
            volume={1},
            year={1989},
            publisher={Pantheon}
            }
        <BLANKLINE>
        ## Evolution
            @book{darwin1859origin,
            title={On the origin of species},
            author={Darwin, Charles},
            year={1859},
            publisher={Lulu. com}
            }
        """  # noqa: B950
        header = ["# Citations"]
        body = []
        for cls in self._registered:
            body.append(self.format_citation(cls))
        return os.linesep.join(header + [(2 * os.linesep).join(body)])

    def __repr__(self):
        return f"ComponentRegistry({repr(self.registered)})"


registry = ComponentRegistry(_info)



import numpy as np

from ..dual import DualGraph
from ..voronoi.dual_voronoi import DualVoronoiGraph
from .hex import (
    HorizontalHexTriGraph,
    HorizontalRectTriGraph,
    TriGraph,
    VerticalHexTriGraph,
    VerticalRectTriGraph,
)


class DualHexGraph(DualGraph, TriGraph):

    """Graph of a structured grid of triangles.

    Examples
    --------
    >>> import numpy as np
    >>> from landlab.graph import DualHexGraph

    >>> graph = DualHexGraph((3, 2), node_layout='hex')
    >>> graph.number_of_nodes
    7
    >>> graph.number_of_corners
    6

    >>> np.round(graph.y_of_node * 2. / np.sqrt(3))
    ...     # doctest: +NORMALIZE_WHITESPACE
    array([ 0.,  0.,  1.,  1.,  1.,  2.,  2.])
    >>> graph.x_of_node # doctest: +NORMALIZE_WHITESPACE
    array([ 0.5,  1.5,  0. ,  1. ,  2. ,  0.5,  1.5])
    """

    def __init__(
        self,
        shape,
        spacing=1.0,
        xy_of_lower_left=(0.0, 0.0),
        orientation="horizontal",
        node_layout="rect",
        sort=False,
    ):
        """Create a structured grid of triangles.

        Parameters
        ----------
        shape : tuple of int
            Number of rows and columns of the hex grid. The first value
            is the number of nodes in the first column and the second the
            number of nodes in the first column.
        spacing : float, optional
            Length of links.
        xy_of_lower_left : tuple of float, optional
            Coordinates of lower-left corner of the grid.
        orientation: {'horizontal', 'vertical'}
            Specify if triangles should be laid out in rows or columns.
        node_layout: {'rect', 'hex'}
            Specify the overall layout of the nodes. Use *rect* for
            the layout to approximate a rectangle and *hex* for
            a hexagon.
        """
        if node_layout not in ("rect", "hex"):
            raise ValueError("node_layout not understood")

        if orientation not in ("horizontal", "vertical"):
            raise ValueError("orientation not understood")

        layouts = {
            "horizontal_hex": HorizontalHexTriGraph,
            "vertical_hex": VerticalHexTriGraph,
            "horizontal_rect": HorizontalRectTriGraph,
            "vertical_rect": VerticalRectTriGraph,
        }
        layout = layouts["_".join([orientation, node_layout])]

        try:
            spacing = float(spacing)
        except TypeError as exc:
            raise TypeError("spacing must be a float") from exc

        self._shape = tuple(shape)
        self._spacing = spacing
        self._orientation = orientation
        self._node_layout = node_layout

        x_of_node, y_of_node = layout.xy_of_node(
            shape, spacing=spacing, xy_of_lower_left=xy_of_lower_left
        )
        self._perimeter_nodes = layout.perimeter_nodes(shape)

        perimeter_links = np.empty((len(self._perimeter_nodes), 2), dtype=int)
        perimeter_links[:, 0] = self._perimeter_nodes
        perimeter_links[:-1, 1] = self._perimeter_nodes[1:]
        perimeter_links[-1, 1] = self._perimeter_nodes[0]

        DualVoronoiGraph.__init__(
            self, (y_of_node, x_of_node), perimeter_links=perimeter_links, sort=False
        )

        if sort:
            self.sort()



import warnings

import numpy as np

from landlab.components.network_sediment_transporter.sediment_pulser_base import (
    SedimentPulserBase,
)
from landlab.data_record import DataRecord

_OUT_OF_NETWORK = -2


class SedimentPulserEachParcel(SedimentPulserBase):

    """Send pulses of sediment to specific point locations within the channel
    network and divide the pulses into parcels. Pulses may be any volume.
    Parcels must be less than or equal to a user specified maximum volume.

    SedimentPulserEachParcel is instantiated by specifying the network model grid
    it will pulse the parcels into

    SedimentPulserEachParcel is run (adds parcels to DataRecrod) by calling the
    SedimentPulserEachParcel instance with the time that pulses are added to
    the channel network and a sediment pulse table (PulseDF)

    PulseDF is a pandas dataframe. At a minimum, the dataframe must have columns 'Link#'
    'normalized_downstream_distance' and 'pulse_volume'. Optionally, the parcel
    volume that the pulse is divided into and grain characteristics of each pulse
    can also be specified in PulseDF.


    .. codeauthor: Jeff Keck, Allison Pfeiffer, Shelby Ahrendt
                   (with help from Eric Hutton and Katy Barnhart)


    Examples
    --------
    >>> import numpy as np
    >>> import pandas as pd
    >>> from landlab import NetworkModelGrid

    Create the network model grid. Pulses are added to the links of the network
    model grid.

    >>> y_of_node = (0, 100, 200, 200, 300, 400, 400, 125)
    >>> x_of_node = (0, 0, 100, -50, -100, 50, -150, -100)
    >>> nodes_at_link = ((1, 0), (2, 1), (1, 7), (3, 1), (3, 4), (4, 5), (4, 6))
    >>> grid = NetworkModelGrid((y_of_node, x_of_node), nodes_at_link)
    >>> grid.at_link["channel_width"] = np.full(grid.number_of_links, 1.0)  # m
    >>> grid.at_link["channel_slope"] = np.full(grid.number_of_links, .01)  # m / m
    >>> grid.at_link["reach_length"] = np.full(grid.number_of_links, 100.0)  # m


    Instantiate 'SedimentPulserEachParcel'

    >>> make_pulse = SedimentPulserEachParcel(grid)

    Define the PulseDF and time of the pulse

    >>> PulseDF = pd.DataFrame(
    ...     {
    ...         "pulse_volume": [0.2, 1, 1.1, 0.5],
    ...         "link_#": [1, 3, 5, 2],
    ...         "normalized_downstream_distance": [0.8, 0.7, 0.5, 0.2],
    ...     }
    ... )
    >>> time = 7

    Run the instance

    >>> parcels = make_pulse(time, PulseDF)

    This should yield a UserWarning: Parcels not provided, created a new DataRecord

    check element_id of each parcel

    >>> print(parcels.dataset['element_id'].values)
    [[1]
    [3]
    [3]
    [5]
    [5]
    [5]
    [2]]

    """

    _name = "SedimentPulserEachParcel"

    _unit_agnostic = False

    _info = {}  # works with the DataRecord

    def __init__(
        self,
        grid,
        parcels=None,
        D50=0.05,
        D84_D50=2.1,
        rho_sediment=2650.0,
        parcel_volume=0.5,
        abrasion_rate=0.0,
    ):
        """
        instantiate SedimentPulserEachParcel

        Parameters
        ----------
        grid : ModelGrid
            landlab *ModelGrid* to place sediment parcels on.
        parcels: landlab DataRecord, optional
            Tracks parcel location and attributes
        D50: float, optional
            median grain size [m]
        D84_D50: float, optional
            ratio of 84th percentile grain size to the median grain size
        rho_sediment : float, optional
            Sediment grain density [kg/m^3].
        parcel_volume : float, optional
            parcel volume [m^3]
        abrasion_rate: float, optional
            volumetric abrasion exponent [1/m]
        """
        SedimentPulserBase.__init__(
            self,
            grid,
            parcels=parcels,
            D50=D50,
            D84_D50=D84_D50,
            rho_sediment=rho_sediment,
            parcel_volume=parcel_volume,
            abrasion_rate=abrasion_rate,
        )

    def __call__(self, time, PulseDF=None):
        """specify the location and attributes of each pulse of material added to
        a Network Model Grid DataRecord

        Parameters
        ----------
        time : integer or datetime64 value equal to nst.time
            time that the pulse is triggered in the network sediment transporter
        PulseDF : pandas dataframe
            each row contains information on the deposition location and volume of
            a single pulse of sediment. The pulse is divided into 'n' number of
            parcels, where 'n' equals the np.ceil(pulse volume / parcel volume)
            For details on the format of the DataFrame, see the docstring for
            function _sediment_pulse_dataframe


        Returns
        -------
            self._parcels
                a DataRecord containing all information on each individual parcel
        """
        # If no PulseDF provided, raise error. Should at least provide an empty PulseDF
        if PulseDF is None:
            raise ValueError("PulseDF was not specified")

        if (
            PulseDF.empty is True
        ):  # if empty, pulser stops, returns the existing parcels, call stops
            warnings.warn("Pulse DataFrame is EMPTY", stacklevel=2)
            return self._parcels

        variables, items = self._sediment_pulse_dataframe(
            time,  # create variabels and and items needed to create the data record
            PulseDF,
        )

        if self._parcels is None:  # if no parcels, create parcels
            self._parcels = DataRecord(
                self._grid,
                items=items,
                time=[time],
                data_vars=variables,
                dummy_elements={"link": [_OUT_OF_NETWORK]},
            )

            warnings.warn(
                "Parcels not provided, created a new DataRecord", stacklevel=2
            )

        else:  # else use the add item method to add parcels
            self._parcels.add_item(time=[time], new_item=items, new_item_spec=variables)

        return self._parcels

    def _sediment_pulse_dataframe(self, time, PulseDF):
        """Convert PulseDF to a :class:`~.DataRecord` formatted for the
        :class:`~.NetworkSedimentTransporter`.

        Parameters
        ----------
        time : integer or datetime64

        PulseDF : pandas dataframe

            The PulseDF must include the following columns:
                'link_#', 'pulse_volume', 'normalized_downstream_distance'

            Optionally, PulseDF can include the following columns:
               'D50', 'D84_D50', 'abrasion_rate', 'rho_sediment', 'parcel_volume'

            Values in each columne are defined as follows:

            'link_#': int - link number pulse enters the channel network
            'pulse_volume: float - total volume of the pulse [m^3]
            'normalized_downstream_distance': float - distance from link inlet
                                                      divided by link length
            'D50': float - median grain size [m]
            'D84_D50': float - grain-size standard deviation [m]
            'abrasion_rate': float - rate that grain size decreases with
                                     distance along channel [mm/km?]
            'rho_sediment': float - density grains [kg/m^3]
            'parcel_volume': float - maximum volume of one parcel [m^3]


            if the optional columns are not included in PulseDF, those parameters
            are assumed uniform across the basin, constant with time and equal
            to the corrisponding class variables.

        Returns
        -------
        tuple: (variables, items)
            variables: dictionary, attribute values for all new parcels
            item_id: dictionary, model grid element and index of element of each parcel

        """
        # split pulse into parcels.
        p_np = []  # list of number of parcels in each pulse
        volume = np.array([])  # list of parcel volumes from all pulses
        for _index, row in PulseDF.iterrows():
            # set the maximum allowable parcel volume using either
            # the default value or value in PulseDF
            if "parcel_volume" in PulseDF:
                mpv = row["parcel_volume"]
            else:
                mpv = self._parcel_volume

            # split the pulse into parcels
            if row["pulse_volume"] < mpv:
                # only one partial parcel volume
                v_p = np.array([row["pulse_volume"]])
            else:
                # number of whole parcels
                n_wp = int(np.floor(row["pulse_volume"] / mpv))
                # array of volumes, whole parcels
                v_wp = np.ones(n_wp) * mpv
                # volume of last parcel, a partial parcel
                v_pp = np.array([row["pulse_volume"] % mpv])
                # array of all parcel volumes
                # partial parcel included if volume > 0
                if v_pp > 0:
                    v_p = np.concatenate((v_wp, v_pp))
                else:
                    v_p = v_wp
            volume = np.concatenate((volume, v_p))
            p_np.append(len(v_p))
        volume = np.expand_dims(volume, axis=1)

        # link location
        link_distance_ratio = np.array([])
        for i, val in enumerate(PulseDF["normalized_downstream_distance"].values):
            # parcels from the same pulse enter channel at the same point
            link_distance_ratio = np.concatenate(
                (link_distance_ratio, np.ones(p_np[i]) * val)
            )
        location_in_link = np.expand_dims(link_distance_ratio, axis=1)

        # element id and starting link
        element_id = np.array([])
        for i, row in PulseDF.iterrows():
            element_id = np.concatenate((element_id, np.ones(p_np[i]) * row["link_#"]))
        starting_link = element_id.copy()
        element_id = np.expand_dims(element_id.astype(int), axis=1)

        # specify that parcels are in the links of the network model grid
        grid_element = ["link"] * np.size(element_id)
        grid_element = np.expand_dims(grid_element, axis=1)

        # time of arrivial (time instance called)
        time_arrival_in_link = np.full(np.shape(element_id), time, dtype=float)

        # All parcels in pulse are in the active layer (1) rather than subsurface (0)
        active_layer = np.ones(np.shape(element_id))

        if "rho_sediment" in PulseDF.columns:
            density = np.array([])
            for i, row in PulseDF.iterrows():
                density = np.concatenate(
                    (density, np.ones(p_np[i]) * row["rho_sediment"])
                )
        else:
            density = self._rho_sediment * np.ones(np.shape(starting_link))

        if "abrasion_rate" in PulseDF.columns:
            abrasion_rate = np.array([])
            for i, row in PulseDF.iterrows():
                abrasion_rate = np.concatenate(
                    (abrasion_rate, np.ones(p_np[i]) * row["abrasion_rate"])
                )
        else:
            abrasion_rate = self._abrasion_rate * np.ones(np.shape(starting_link))

        if "D50" in PulseDF.columns and "D84_D50" in PulseDF.columns:
            grain_size = np.array([])
            for i, row in PulseDF.iterrows():
                # det D50 and D84_D50
                n_parcels = p_np[i]
                D50 = row["D50"]
                D84_D50 = row["D84_D50"]
                grain_size_pulse = np.random.lognormal(
                    np.log(D50), np.log(D84_D50), n_parcels
                )
                grain_size = np.concatenate((grain_size, grain_size_pulse))
        else:
            n_parcels = sum(p_np)
            D50 = self._D50
            D84_D50 = self._D84_D50
            grain_size = np.random.lognormal(np.log(D50), np.log(D84_D50), n_parcels)

        grain_size = np.expand_dims(grain_size, axis=1)

        return {
            "starting_link": (["item_id"], starting_link),
            "abrasion_rate": (["item_id"], abrasion_rate),
            "density": (["item_id"], density),
            "time_arrival_in_link": (["item_id", "time"], time_arrival_in_link),
            "active_layer": (["item_id", "time"], active_layer),
            "location_in_link": (["item_id", "time"], location_in_link),
            "D": (["item_id", "time"], grain_size),
            "volume": (["item_id", "time"], volume),
        }, {"grid_element": grid_element, "element_id": element_id}



import numpy as np
import pytest

from landlab import HexModelGrid


def test_perimeter_nodes():
    """Test perimeter nodes of a hex grid."""
    grid = HexModelGrid((3, 4), node_layout="rect")
    assert np.all(grid.perimeter_nodes == [3, 7, 11, 10, 9, 8, 4, 0, 1, 2])


def test_right_edge_nodes():
    """Test right edge nodes of a hex grid."""
    grid = HexModelGrid((3, 4), node_layout="rect")
    assert np.all(grid.nodes_at_right_edge == [3, 7, 11])


def test_top_edge_nodes():
    """Test top edge nodes of a hex grid."""
    grid = HexModelGrid((3, 4), node_layout="rect")
    assert np.all(grid.nodes_at_top_edge == [8, 9, 10, 11])


def test_left_edge_nodes():
    """Test left edge nodes of a hex grid."""
    grid = HexModelGrid((3, 4), node_layout="rect")
    assert np.all(grid.nodes_at_left_edge == [0, 4, 8])


def test_bottom_edge_nodes():
    """Test bottom edge nodes of a hex grid."""
    grid = HexModelGrid((3, 4), node_layout="rect")
    assert np.all(grid.nodes_at_bottom_edge == [0, 1, 2, 3])


def test_edges_are_readonly(edge_name):
    grid = HexModelGrid((3, 4), node_layout="rect")
    assert not grid.perimeter_nodes.flags["WRITEABLE"]
    with pytest.raises(ValueError):
        getattr(grid, "nodes_at_" + edge_name)[0] = 999


def test_edges_are_cached(edge_name):
    grid = HexModelGrid((3, 4), node_layout="rect")
    x = grid.perimeter_nodes
    assert grid.perimeter_nodes is x
    x = getattr(grid, "nodes_at_" + edge_name)
    assert getattr(grid, "nodes_at_" + edge_name) is x



from landlab import Component


class AreaSlopeTransporter(Component):
    """Model drainage network evolution for a network of transport-limited
    rivers in which sediment transport rate is calculated as a power-law
    function of drainage area and local streamwise slope gradient.

    AreaSlopeTransporter is designed to operate together with a flow-routing
    component such as PriorityFloodFlowRouter, so that each grid node has
    a defined flow direction toward one of its neighbor nodes. Each core node
    is assumed to contain one outgoing fluvial channel, and (depending on
    the drainage structure) zero, one, or more incoming channels. These channels are
    treated as effectively sub-grid-scale features that are embedded in valleys
    that have a width of one grid cell. The rate of sediment transport out of
    a given node is calculated as a generic power function of drainage area,
    local slope, and a user-specified transport coefficient.
    Similar power-law formulations have been used, for example, by
    Willgoose et al. (1991a,b,c, and many papers following that use the
    SIBERIA model) and Howard (1994, in Water Resources Research).

    Parameters
    ----------
    grid : ModelGrid
        A Landlab model grid object
    transport_coefficient : float (default 0.0055)
        Dimensional transport efficiency factor
    area_exponent : float (default 1.4)
        Exponent on effective total discharge
    slope_exponent : float (default 2.1)
        Exponent on local streamwise slope gradient

    Examples
    --------
    >>> from landlab import RasterModelGrid
    >>> from landlab.components import FlowAccumulator
    >>> grid = RasterModelGrid((3, 3), xy_spacing=1000.0)
    >>> elev = grid.add_zeros("topographic__elevation", at="node")
    >>> grid.status_at_node[grid.perimeter_nodes] = grid.BC_NODE_IS_CLOSED
    >>> grid.status_at_node[5] = grid.BC_NODE_IS_FIXED_VALUE
    >>> fa = FlowAccumulator(grid)
    >>> fa.run_one_step()
    >>> transporter = AreaSlopeTransporter(grid)
    >>> for _ in range(200):
    ...     fa.run_one_step()
    ...     elev[grid.core_nodes] += 1.0
    ...     transporter.run_one_step(10000.0)
    >>> int(round(elev[4] * 100))
    1068
    """

    _name = "AreaSlopeTransporter"

    _unit_agnostic = True

    _info = {
        "sediment__volume_influx": {
            "dtype": float,
            "intent": "out",
            "optional": False,
            "units": "m**3/y",
            "mapping": "node",
            "doc": "Volumetric incoming streamwise sediment transport rate",
        },
        "sediment__volume_outflux": {
            "dtype": float,
            "intent": "out",
            "optional": False,
            "units": "m**3/y",
            "mapping": "node",
            "doc": "Volumetric outgoing streamwise sediment transport rate",
        },
        "flow__link_to_receiver_node": {
            "dtype": int,
            "intent": "in",
            "optional": False,
            "units": "-",
            "mapping": "node",
            "doc": "ID of link downstream of each node, which carries the discharge",
        },
        "flow__receiver_node": {
            "dtype": int,
            "intent": "in",
            "optional": False,
            "units": "-",
            "mapping": "node",
            "doc": "Node array of receivers (node that receives flow from current node)",
        },
        "flow__upstream_node_order": {
            "dtype": int,
            "intent": "in",
            "optional": False,
            "units": "-",
            "mapping": "node",
            "doc": "Node array containing downstream-to-upstream ordered list of node IDs",
        },
        "sediment__rate_of_change": {
            "dtype": float,
            "intent": "out",
            "optional": False,
            "units": "m/y",
            "mapping": "node",
            "doc": "Time rate of change of sediment thickness",
        },
        "drainage_area": {
            "dtype": float,
            "intent": "in",
            "optional": False,
            "units": "m**2",
            "mapping": "node",
            "doc": "Upstream accumulated surface area contributing to the node's discharge",
        },
        "topographic__elevation": {
            "dtype": float,
            "intent": "inout",
            "optional": False,
            "units": "m",
            "mapping": "node",
            "doc": "Land surface topographic elevation",
        },
        "topographic__steepest_slope": {
            "dtype": float,
            "intent": "in",
            "optional": False,
            "units": "-",
            "mapping": "node",
            "doc": "The steepest *downhill* slope",
        },
    }

    def __init__(
        self,
        grid,
        transport_coefficient=0.0055,
        area_exponent=1.4,
        slope_exponent=2.1,
    ):
        """Initialize AreaSlopeTransporter."""

        super().__init__(grid)

        # Parameters
        self._trans_coef = transport_coefficient
        self._area_exponent = area_exponent
        self._slope_exponent = slope_exponent

        # Fields and arrays
        self._elev = grid.at_node["topographic__elevation"]
        self._area = grid.at_node["drainage_area"]
        self._slope = grid.at_node["topographic__steepest_slope"]
        self._receiver_node = grid.at_node["flow__receiver_node"]
        self._receiver_link = grid.at_node["flow__link_to_receiver_node"]
        super().initialize_output_fields()
        self._sediment_influx = grid.at_node["sediment__volume_influx"]
        self._sediment_outflux = grid.at_node["sediment__volume_outflux"]
        self._dzdt = grid.at_node["sediment__rate_of_change"]

    def calc_transport_capacity(self):
        """Calculate and return bed-load transport capacity.

        Calculation uses power-law approach, and provides
        volume per time rate.

        Examples
        --------
        >>> from landlab import RasterModelGrid
        >>> from landlab.components import FlowAccumulator
        >>> grid = RasterModelGrid((3, 3), xy_spacing=100.0)
        >>> elev = grid.add_zeros("topographic__elevation", at="node")
        >>> elev[3:] = 1.0
        >>> fa = FlowAccumulator(grid)
        >>> fa.run_one_step()
        >>> transporter = AreaSlopeTransporter(grid)
        >>> transporter.calc_transport_capacity()
        >>> int(transporter._sediment_outflux[4] * 1000)
        138
        """
        self._sediment_outflux[:] = (
            self._trans_coef
            * self._area**self._area_exponent
            * self._slope**self._slope_exponent
        )

    def calc_sediment_rate_of_change(self):
        """Update the rate of thickness change of sediment at each core node.

        Examples
        --------
        >>> import numpy as np
        >>> from landlab import RasterModelGrid
        >>> from landlab.components import FlowAccumulator
        >>> grid = RasterModelGrid((3, 4), xy_spacing=100.0)
        >>> elev = grid.add_zeros("topographic__elevation", at="node")
        >>> elev[:] = 0.01 * grid.x_of_node
        >>> grid.status_at_node[grid.perimeter_nodes] = grid.BC_NODE_IS_CLOSED
        >>> grid.status_at_node[4] = grid.BC_NODE_IS_FIXED_VALUE
        >>> fa = FlowAccumulator(grid)
        >>> fa.run_one_step()
        >>> transporter = AreaSlopeTransporter(grid)
        >>> transporter.calc_sediment_rate_of_change()
        >>> np.round(transporter._sediment_outflux[4:7], 3)
        array([ 0.   ,  0.365,  0.138])
        >>> np.round(transporter._sediment_influx[4:7], 3)
        array([ 0.365,  0.138,  0.   ])
        >>> np.round(transporter._dzdt[5:7], 8)
        array([ -2.26400000e-05,  -1.38200000e-05])
        """
        self.calc_transport_capacity()
        cores = self.grid.core_nodes
        self._sediment_influx[:] = 0.0
        for c in cores:  # send sediment downstream
            r = self._receiver_node[c]
            self._sediment_influx[r] += self._sediment_outflux[c]
        self._dzdt[cores] = (
            self._sediment_influx[cores] - self._sediment_outflux[cores]
        ) / self.grid.area_of_cell[self.grid.cell_at_node[cores]]

    def run_one_step(self, dt):
        """Advance solution by time interval dt.

        Examples
        --------
        >>> import numpy as np
        >>> from landlab import RasterModelGrid
        >>> from landlab.components import FlowAccumulator
        >>> grid = RasterModelGrid((3, 4), xy_spacing=100.0)
        >>> elev = grid.add_zeros("topographic__elevation", at="node")
        >>> elev[:] = 0.01 * grid.x_of_node
        >>> grid.status_at_node[grid.perimeter_nodes] = grid.BC_NODE_IS_CLOSED
        >>> grid.status_at_node[4] = grid.BC_NODE_IS_FIXED_VALUE
        >>> fa = FlowAccumulator(grid)
        >>> fa.run_one_step()
        >>> transporter = AreaSlopeTransporter(grid)
        >>> transporter.run_one_step(10000.0)
        >>> np.round(elev[4:7], 4)
        array([ 0.    ,  0.7736,  1.8618])
        """
        self.calc_sediment_rate_of_change()
        self._elev += self._dzdt * dt



import pathlib

import numpy as np
import xarray as xr


def to_netcdf(
    grid, path, include="*", exclude=None, time=None, format="NETCDF4", mode="w"
):
    """Write landlab a grid to a netcdf file.

    Write the data and grid information for *grid* to *path* as NetCDF.
    If the *append* keyword argument in True, append the data to an existing
    file, if it exists. Otherwise, clobber an existing files.

    Parameters
    ----------
    grid : ModelGrid
        Landlab grid object that holds a grid and field values.
    path : str
        Path to which to save this grid.
    include : str or iterable of str, optional
        A list of unix-style glob patterns of field names to include. Fully
        qualified field names that match any of these patterns will be
        written to the output file. A fully qualified field name is one that
        that has a prefix that indicates what grid element is defined on
        (e.g. "at_node:topographic__elevation"). The default is to include
        all fields.
    exclude : str or iterable of str, optional
        Like the *include* keyword but, instead, fields matching these
        patterns will be excluded from the output file.
    format : {'NETCDF3_CLASSIC', 'NETCDF3_64BIT', 'NETCDF4_CLASSIC', 'NETCDF4'}
        Format of output netcdf file.
    attrs : dict
        Attributes to add to netcdf file.
    mode : {"w", "a"}, optional
        Write ("w") or append ("a") mode. If mode="w", any existing file at
        this location will be overwritten. If mode="a", existing variables
        will be overwritten.


    Parameters
    ----------
    Examples
    --------
    >>> import numpy as np
    >>> from landlab import RasterModelGrid
    >>> from landlab.io.netcdf import to_netcdf

    Create a uniform rectilinear grid with four rows and 3 columns, and add
    some data fields to it.

    >>> rmg = RasterModelGrid((4, 3))
    >>> rmg.at_node["topographic__elevation"] = np.arange(12.0)
    >>> rmg.at_node["uplift_rate"] = 2.0 * np.arange(12.0)

    Create a temporary directory to write the netcdf file into.

    >>> import tempfile, os
    >>> temp_dir = tempfile.mkdtemp()
    >>> os.chdir(temp_dir)

    Write the grid to a netcdf3 file but only include the *uplift_rate*
    data in the file.

    >>> to_netcdf(
    ...     rmg, "test.nc", format="NETCDF3_64BIT", include="at_node:uplift_rate"
    ... )

    Read the file back in and check its contents.

    >>> from scipy.io import netcdf
    >>> fp = netcdf.netcdf_file('test.nc', 'r')
    >>> 'at_node:uplift_rate' in fp.variables
    True
    >>> 'at_node:topographic__elevation' in fp.variables
    False
    >>> fp.variables['at_node:uplift_rate'][:].flatten()
    array([  0.,   2.,   4.,   6.,   8.,  10.,  12.,  14.,  16.,  18.,  20.,
            22.])

    >>> rmg.at_cell["air__temperature"] = np.arange(2.0)
    >>> to_netcdf(
    ...     rmg,
    ...     "test-cell.nc",
    ...     format="NETCDF3_64BIT",
    ...     include="at_cell:*",
    ...     # names="air__temperature", at="cell",
    ... )
    """
    path = pathlib.Path(path)
    if not path.is_file():
        mode = "w"

    if time is None and mode == "a":
        time = np.nan

    this_dataset = grid.as_dataset(include=include, exclude=exclude, time=time)

    if format != "NETCDF4":
        this_dataset["status_at_node"] = (
            ("node",),
            this_dataset["status_at_node"].values.astype(dtype=int),
        )

    if mode == "a":
        with xr.open_dataset(path) as that_dataset:
            if "time" not in that_dataset.dims:
                _add_time_dimension_to_dataset(that_dataset, time=np.nan)

            new_vars = set(this_dataset.variables) - set(that_dataset.variables)
            for var in new_vars:
                that_dataset[var] = (
                    this_dataset[var].dims,
                    np.full_like(this_dataset[var].values, np.nan),
                )

            for var in list(that_dataset.variables):
                if var.startswith("at_layer"):
                    del that_dataset[var]

            this_dataset = xr.concat(
                [that_dataset, this_dataset], dim="time", data_vars="minimal"
            )

            if np.isnan(this_dataset["time"][-1]):
                this_dataset["time"].values[-1] = this_dataset["time"][-2] + 1.0

    this_dataset.to_netcdf(path, format=format, mode="w", unlimited_dims=("time",))


def _add_time_dimension_to_dataset(dataset, time=0.0):
    """Add a time dimension to all variables except those at_layer."""
    names = {
        name
        for name in dataset.variables
        if name.startswith("at_") and not name.startswith("at_layer")
    }

    for name in names:
        dataset[name] = (("time",) + dataset[name].dims, dataset[name].values[None])
    dataset["time"] = (("time",), [time])



#!/usr/bin/env python
"""
.. codeauthor:: D Litwin, G Tucker

.. sectionauthor:: D Litwin, G Tucker
"""

from .dupuit_percolator import GroundwaterDupuitPercolator

__all__ = ["GroundwaterDupuitPercolator"]



# distutils: language = c++
import numpy as np
cimport numpy as cnp
cimport cython
from libcpp cimport bool
from libcpp.pair cimport pair

cdef extern from "_priority_queue.hpp" nogil:
    cdef cppclass _priority_queue:
        _priority_queue(...) except +
        void push(pair[cnp.int_t, cnp.float_t])
        pair[cnp.int_t, cnp.float_t] top() except +
        void pop()
        bool empty()
        cnp.int_t size()



# distutils: language = c++
import numpy as np
cimport numpy as cnp
cimport cython
from libcpp cimport bool
from libcpp.pair cimport pair

cdef extern from "_priority_queue.hpp" nogil:
    cdef cppclass _priority_queue:
        _priority_queue(...) except +
        void push(pair[cnp.int_t, cnp.float_t])
        pair[cnp.int_t, cnp.float_t] top() except +
        void pop()
        bool empty()
        cnp.int_t size()

cdef bool _compare_second(pair[int, double] a, pair[int, double] b) nogil

cdef void _init_flow_direction_queues(
    const cnp.int_t [:] base_level_nodes, const cnp.int_t [:] closed_nodes,
    cnp.float_t [:] z, _priority_queue& to_do, cnp.int_t [:] receivers,
    cnp.int_t [:] outlet_nodes, cnp.int_t [:] done, cnp.int_t* done_n_ptr) nogil

cdef void _set_flooded_and_outlet(cnp.int_t donor_id, cnp.float_t [:] z,
    cnp.int_t [:] receivers, cnp.int_t [:] outlet_nodes,
    cnp.int_t [:] depression_outlet_nodes, cnp.int_t [:] flooded_nodes,
    cnp.float_t [:] depression_depths, cnp.float_t [:] depression_free_elevations,
    cnp.int_t flooded_status, cnp.int_t bad_index,
    cnp.float_t min_elevation_relative_diff) nogil

cdef void _set_receiver(cnp.int_t donor_id, cnp.int_t receiver_id,
    cnp.int_t [:] receivers, cnp.int_t [:] done, cnp.int_t* done_n_ptr) nogil

cdef void _set_donor_properties(cnp.int_t donor_id, cnp.int_t receiver_id,
    cnp.int_t [:] sorted_pseudo_tails, const cnp.int_t [:,:] head_start_end_indexes,
    const cnp.int_t [:] sorted_dupli_links,
    cnp.float_t [:] sorted_dupli_gradients, cnp.float_t [:] z,
    cnp.float_t [:] steepest_slopes, cnp.int_t [:] links_to_receivers) nogil

#######################################################################################

cdef void _direct_flow_c(cnp.int_t nodes_n, const cnp.int_t[:] base_level_nodes,
    const cnp.int_t[:] closed_nodes, cnp.int_t[:] sorted_pseudo_tails,
    cnp.float_t[:] sorted_dupli_gradients, const cnp.int_t[:] sorted_dupli_links,
    const cnp.int_t[:, :] head_start_end_indexes, cnp.int_t [:] outlet_nodes,
    cnp.int_t [:] depression_outlet_nodes, cnp.int_t[:] flooded_nodes,
    cnp.float_t[:] depression_depths, cnp.float_t [:] depression_free_elevations,
    cnp.int_t[:] links_to_receivers, cnp.int_t[:] receivers,
    cnp.float_t[:] steepest_slopes, cnp.float_t[:] z, cnp.int_t flooded_status,
    cnp.int_t bad_index, cnp.int_t neighbors_max_number,
    cnp.float_t min_elevation_relative_diff)



import numpy as np
from numpy.testing import assert_array_almost_equal

from landlab.components import NetworkSedimentTransporter
from landlab.data_record import DataRecord


def test_grainsize_active_layer(example_nmg, example_parcels, example_flow_director):
    time = [0.0]  # probably not the sensible way to do this...

    items = {"grid_element": "link", "element_id": np.array([[6], [6]])}

    initial_volume = np.array([[1], [1]])
    abrasion_rate = np.array([0, 0])

    variables = {
        "starting_link": (["item_id"], np.array([6, 6])),
        "abrasion_rate": (["item_id"], abrasion_rate),
        "density": (["item_id"], np.array([2650, 2650])),
        "time_arrival_in_link": (["item_id", "time"], np.array([[0.5], [0]])),
        "active_layer": (["item_id", "time"], np.array([[1], [1]])),
        "location_in_link": (["item_id", "time"], np.array([[0], [0]])),
        "D": (["item_id", "time"], np.array([[0.05], [0.05]])),
        "volume": (["item_id", "time"], initial_volume),
    }

    two_parcels = DataRecord(
        example_nmg,
        items=items,
        time=time,
        data_vars=variables,
        dummy_elements={"link": [NetworkSedimentTransporter.OUT_OF_NETWORK]},
    )

    timesteps = 8

    example_nmg.at_link["flow_depth"] = (
        example_nmg.at_link["flow_depth"] * 5
    )  # high transport rate

    nst = NetworkSedimentTransporter(
        example_nmg,
        two_parcels,
        example_flow_director,
        bed_porosity=0.03,
        g=9.81,
        fluid_density=1000,
        transport_method="WilcockCrowe",
        active_layer_method="GrainSizeDependent",
        active_layer_d_multiplier=3,
    )

    dt = 60 * 60 * 24  # (seconds) daily timestep

    for _ in range(0, (timesteps * dt), dt):
        nst.run_one_step(dt)

    # Active layer thickness should be 3*d_mean
    active_layer_thickness_should_be = 3 * 0.05

    assert_array_almost_equal(
        active_layer_thickness_should_be, nst._active_layer_thickness[0]
    )


def test_constant_active_layer(example_nmg, example_parcels, example_flow_director):
    time = [0.0]  # probably not the sensible way to do this...

    items = {"grid_element": "link", "element_id": np.array([[6], [6]])}

    initial_volume = np.array([[1], [1]])
    abrasion_rate = np.array([0, 0])

    variables = {
        "starting_link": (["item_id"], np.array([6, 6])),
        "abrasion_rate": (["item_id"], abrasion_rate),
        "density": (["item_id"], np.array([2650, 2650])),
        "time_arrival_in_link": (["item_id", "time"], np.array([[0.5], [0]])),
        "active_layer": (["item_id", "time"], np.array([[1], [1]])),
        "location_in_link": (["item_id", "time"], np.array([[0], [0]])),
        "D": (["item_id", "time"], np.array([[0.05], [0.05]])),
        "volume": (["item_id", "time"], initial_volume),
    }

    two_parcels = DataRecord(
        example_nmg,
        items=items,
        time=time,
        data_vars=variables,
        dummy_elements={"link": [NetworkSedimentTransporter.OUT_OF_NETWORK]},
    )

    timesteps = 8

    example_nmg.at_link["flow_depth"] = (
        example_nmg.at_link["flow_depth"] * 5
    )  # high transport rate

    nst = NetworkSedimentTransporter(
        example_nmg,
        two_parcels,
        example_flow_director,
        bed_porosity=0.03,
        g=9.81,
        fluid_density=1000,
        transport_method="WilcockCrowe",
        active_layer_method="Constant10cm",
    )

    dt = 60 * 60 * 24  # (seconds) daily timestep

    for _ in range(0, (timesteps * dt), dt):
        nst.run_one_step(dt)

    # Active layer thickness should be 10 cm
    active_layer_thickness_should_be = 0.10

    assert_array_almost_equal(
        active_layer_thickness_should_be, nst._active_layer_thickness[0]
    )



import copy as cp

import numpy as np
import pytest
from numpy import testing

from landlab import FieldError, HexModelGrid, RasterModelGrid
from landlab.components import (
    FlowAccumulator,
    PriorityFloodFlowRouter,
    SpaceLargeScaleEroder,
)

try:
    PriorityFloodFlowRouter.load_richdem()
except ModuleNotFoundError:
    with_richdem = False
else:
    with_richdem = True


def test_inputFields_flowRouter():
    """
    SpaceLargeScaleEroder should throw an error when topograhy is not equal to the sum of
    bedrock and soil thickness
    """
    # %%
    # Make a raster model grid and create a plateau
    mg = RasterModelGrid((5, 5))
    z = mg.add_zeros("topographic__elevation", at="node")
    br = mg.add_zeros("bedrock__elevation", at="node")
    br[:] = mg.x_of_node + mg.y_of_node
    soil = mg.add_zeros("soil__depth", at="node")
    z[:] = br + soil
    fa = FlowAccumulator(mg, flow_director="D8")
    fa.run_one_step()

    # make plateau at 10m
    br += 10

    # Instanciate the slider
    with pytest.raises(AssertionError):
        _ = SpaceLargeScaleEroder(mg)


# %%
def test_inputFields_soil():
    """
    SpaceLargeScaleEroder should throw an error when the soil__depth field is not provided
    """
    # %%
    mg = RasterModelGrid((5, 5))
    _ = mg.add_zeros("topographic__elevation", at="node")
    _ = mg.add_zeros("bedrock__elevation", at="node")
    fa = FlowAccumulator(mg, flow_director="D8")
    fa.run_one_step()

    # Instanciate the slider
    with pytest.raises(FieldError):
        _ = SpaceLargeScaleEroder(mg)


# %%
def test_inputFields_bedrock():
    """
    SpaceLargeScaleEroder should instanciate the bedrock__elevation field
    when it is not provided
    """
    # %%
    mg = RasterModelGrid((5, 5))
    _ = mg.add_zeros("topographic__elevation", at="node")
    _ = mg.add_zeros("soil__depth", at="node")
    fa = FlowAccumulator(mg, flow_director="D8")
    fa.run_one_step()
    _ = SpaceLargeScaleEroder(mg)

    assert "bedrock__elevation" in mg.at_node


# %%
def test_properties_phi_fraction_fines_LS():
    """
    SpaceLargeScaleEroder should throw an error when phi/fraction_fines_LS < 0 or phi > 0
    """
    # %%
    mg = RasterModelGrid((5, 5))
    z = mg.add_zeros("topographic__elevation", at="node")
    br = mg.add_zeros("bedrock__elevation", at="node")
    br[:] = mg.x_of_node + mg.y_of_node
    soil = mg.add_zeros("soil__depth", at="node")
    z[:] = br + soil
    fa = FlowAccumulator(mg, flow_director="D8")
    fa.run_one_step()

    # Instanciate the slider
    with pytest.raises(ValueError):
        _ = SpaceLargeScaleEroder(mg, phi=-0.2)
    # Instanciate the slider
    with pytest.raises(ValueError):
        _ = SpaceLargeScaleEroder(mg, phi=1.2)
    # Instanciate the slider
    with pytest.raises(ValueError):
        _ = SpaceLargeScaleEroder(mg, F_f=-0.2)
    # Instanciate the slider
    with pytest.raises(ValueError):
        _ = SpaceLargeScaleEroder(mg, F_f=1.2)


# %%


def test_route_to_multiple_error_raised():
    # %%
    mg = RasterModelGrid((10, 10))
    z = mg.add_zeros("topographic__elevation", at="node")
    br = mg.add_zeros("bedrock__elevation", at="node")
    br[:] = mg.x_of_node + mg.y_of_node
    soil = mg.add_zeros("soil__depth", at="node")
    z[:] = br + soil
    fa = FlowAccumulator(mg, flow_director="MFD")
    fa.run_one_step()

    with pytest.raises(NotImplementedError):
        SpaceLargeScaleEroder(
            mg,
            K_sed=0.1,
            K_br=0.1,
            F_f=0.5,
            phi=0.1,
            H_star=1.0,
            v_s=0.001,
            m_sp=1.0,
            n_sp=0.5,
            sp_crit_sed=0,
            sp_crit_br=0,
        )


# %%
def test_soil_field_already_on_grid():
    # %%
    """
    Test that an existing soil grid field is not changed by instantiating
    SpaceLargeScaleEroder.
    """

    # set up a 5x5 grid with one open outlet node and low initial elevations.
    nr = 5
    nc = 5
    mg = RasterModelGrid((nr, nc), xy_spacing=10.0)

    z = mg.add_zeros("topographic__elevation", at="node")
    br = mg.add_zeros("bedrock__elevation", at="node")
    soil = mg.add_zeros("soil__depth", at="node")
    soil += 1.0  # add 1m of soil everywehre

    mg["node"]["topographic__elevation"] += (
        mg.node_y / 10000 + mg.node_x / 10000 + np.random.rand(len(mg.node_y)) / 10000
    )
    mg.set_closed_boundaries_at_grid_edges(
        bottom_is_closed=True,
        left_is_closed=True,
        right_is_closed=True,
        top_is_closed=True,
    )
    mg.set_watershed_boundary_condition_outlet_id(
        0, mg["node"]["topographic__elevation"], -9999.0
    )
    br[:] = z - soil

    # Create a D8 flow handler
    FlowAccumulator(mg, flow_director="D8")

    # Instantiate SpaceLargeScaleEroder
    sp = SpaceLargeScaleEroder(
        mg,
        K_sed=0.01,
        K_br=0.01,
        F_f=0.0,
        phi=0.0,
        v_s=0.001,
        m_sp=0.5,
        n_sp=1.0,
        sp_crit_sed=0,
        sp_crit_br=0,
    )

    # ensure that 'soil__depth' field is everywhere equal to 1.0 m.
    testing.assert_array_equal(
        np.ones(mg.number_of_nodes),
        sp._soil__depth,
        err_msg="SpaceLargeScaleEroder soil depth field test failed",
        verbose=True,
    )

    # %% Check getters
    testing.assert_array_equal(
        0.01,
        sp.K_br,
        err_msg="Parameter value issue",
        verbose=True,
    )
    testing.assert_array_equal(
        0.01,
        sp.K_sed,
        err_msg="Parameter value issue",
        verbose=True,
    )
    # sediment erosion is zero before running the component
    testing.assert_array_equal(
        np.zeros(mg.number_of_nodes),
        sp.Es,
        err_msg="Parameter value issue",
        verbose=True,
    )
    # rock erosion is zero before running the component
    testing.assert_array_equal(
        np.zeros(mg.number_of_nodes),
        sp.Er,
        err_msg="Parameter value issue",
        verbose=True,
    )
    # %% Check setters
    sp.K_br = 0.02
    testing.assert_array_equal(
        0.02,
        sp.K_br,
        err_msg="Parameter value issue",
        verbose=True,
    )
    sp.K_sed = 0.02
    testing.assert_array_equal(
        0.02,
        sp.K_sed,
        err_msg="Parameter value issue",
        verbose=True,
    )

    with pytest.raises(AttributeError):
        sp.Es = np.zeros(mg.number_of_nodes)

    with pytest.raises(AttributeError):
        sp.Er = np.zeros(mg.number_of_nodes)


# %%


def test_br_field_already_on_grid():
    # %%
    """
    Test that an existing bedrock elevation grid field is not changed by
    instantiating SpaceLargeScaleEroder.
    """

    # set up a 5x5 grid with one open outlet node and low initial elevations.
    nr = 5
    nc = 5
    mg = RasterModelGrid((nr, nc), xy_spacing=10.0)

    z = mg.add_zeros("topographic__elevation", at="node")
    br = mg.add_zeros("bedrock__elevation", at="node")
    br += 1.0  # make bedrock elevation 5m below surface
    soil = mg.add_zeros("soil__depth", at="node")

    mg["node"]["topographic__elevation"] += (
        mg.node_y / 10000 + mg.node_x / 10000 + np.random.rand(len(mg.node_y)) / 10000
    )
    mg.set_closed_boundaries_at_grid_edges(
        bottom_is_closed=True,
        left_is_closed=True,
        right_is_closed=True,
        top_is_closed=True,
    )
    mg.set_watershed_boundary_condition_outlet_id(
        0, mg["node"]["topographic__elevation"], -9999.0
    )
    z[:] = br + soil

    # Create a D8 flow handler
    FlowAccumulator(mg, flow_director="D8")

    # Instantiate SpaceLargeScaleEroder
    sp = SpaceLargeScaleEroder(
        mg,
        K_sed=0.01,
        K_br=0.01,
        F_f=0.0,
        phi=0.0,
        v_s=0.001,
        m_sp=0.5,
        n_sp=1.0,
        sp_crit_sed=0,
        sp_crit_br=0,
    )

    # ensure that 'bedrock__elevation' field is everywhere equal to 1.0 m.
    testing.assert_array_equal(
        np.ones(mg.number_of_nodes),
        sp._bedrock__elevation,
        err_msg="SpaceLargeScaleEroder bedrock field test failed",
        verbose=True,
    )


@pytest.mark.parametrize("k_sed", (0.00001, [0.00001] * 25))
def test_matches_detachment_solution(k_sed):
    # %%
    """
    Test that model matches the detachment-limited analytical solution
    for slope/area relationship at steady state: S=(U/K_br)^(1/n)*A^(-m/n).
    """

    # %% set up a 5x5 grid with one open outlet node and low initial elevations.
    nr = 5
    nc = 5
    mg = RasterModelGrid((nr, nc), xy_spacing=10.0)

    z = mg.add_zeros("topographic__elevation", at="node")
    br = mg.add_zeros("bedrock__elevation", at="node")
    soil = mg.add_zeros("soil__depth", at="node")

    mg["node"]["topographic__elevation"] += (
        mg.node_y / 10000 + mg.node_x / 10000 + np.random.rand(len(mg.node_y)) / 10000
    )
    mg.set_closed_boundaries_at_grid_edges(
        bottom_is_closed=True,
        left_is_closed=True,
        right_is_closed=True,
        top_is_closed=True,
    )
    mg.set_watershed_boundary_condition_outlet_id(
        0, mg["node"]["topographic__elevation"], -9999.0
    )
    br[:] = z[:] - soil[:]

    # Create a D8 flow handler
    fa = FlowAccumulator(mg, flow_director="D8")

    # Parameter values for detachment-limited test
    K_br = 0.01
    U = 0.0001
    dt = 1.0
    F_f = 1.0  # all detached rock disappears; detachment-ltd end-member
    m_sp = 0.5
    n_sp = 1.0

    # Instantiate the SpaceLargeScaleEroder component...
    sp = SpaceLargeScaleEroder(
        mg,
        K_sed=k_sed,
        K_br=K_br,
        F_f=F_f,
        phi=0.1,
        H_star=1.0,
        v_s=0.001,
        m_sp=m_sp,
        n_sp=n_sp,
        sp_crit_sed=0,
        sp_crit_br=0,
    )

    # ... and run it to steady state (2000x1-year timesteps).
    for _ in range(2000):
        fa.run_one_step()
        sp.run_one_step(dt=dt)
        z[mg.core_nodes] += U * dt  # m
        br[mg.core_nodes] = z[mg.core_nodes] - soil[mg.core_nodes]

    # compare numerical and analytical slope solutions
    num_slope = mg.at_node["topographic__steepest_slope"][mg.core_nodes]
    analytical_slope = np.power(U / K_br, 1.0 / n_sp) * np.power(
        mg.at_node["drainage_area"][mg.core_nodes], -m_sp / n_sp
    )

    # test for match with analytical slope-area relationship
    testing.assert_array_almost_equal(
        num_slope,
        analytical_slope,
        decimal=8,
        err_msg="SpaceLargeScaleEroder detachment-limited test failed",
        verbose=True,
    )


# %%
def test_matches_detachment_solution_n_gr_1():
    # %%
    """
    Test that model matches the detachment-limited analytical solution
    for slope/area relationship at steady state: S=(U/K_br)^(1/n)*A^(-m/n).
    """

    # %% set up a 5x5 grid with one open outlet node and low initial elevations.
    nr = 5
    nc = 5
    mg = RasterModelGrid((nr, nc), xy_spacing=10.0)

    z = mg.add_zeros("topographic__elevation", at="node")
    br = mg.add_zeros("bedrock__elevation", at="node")
    soil = mg.add_zeros("soil__depth", at="node")

    mg["node"]["topographic__elevation"] += (
        mg.node_y / 10000 + mg.node_x / 10000 + np.random.rand(len(mg.node_y)) / 10000
    )
    mg.set_closed_boundaries_at_grid_edges(
        bottom_is_closed=True,
        left_is_closed=True,
        right_is_closed=True,
        top_is_closed=True,
    )
    mg.set_watershed_boundary_condition_outlet_id(
        0, mg["node"]["topographic__elevation"], -9999.0
    )
    br[:] = z[:] - soil[:]

    # Create a D8 flow handler
    fa = FlowAccumulator(mg, flow_director="D8")

    # Parameter values for detachment-limited test
    K_br = 0.01
    U = 0.0001
    dt = 1.0
    F_f = 1.0  # all detached rock disappears; detachment-ltd end-member
    m_sp = 0.5
    n_sp = 1.1

    # Instantiate the SpaceLargeScaleEroder component...
    sp = SpaceLargeScaleEroder(
        mg,
        K_sed=0.00001,
        K_br=K_br,
        F_f=F_f,
        phi=0.1,
        H_star=1.0,
        v_s=0.001,
        m_sp=m_sp,
        n_sp=n_sp,
        sp_crit_sed=0,
        sp_crit_br=0,
    )

    # ... and run it to steady state (2000x1-year timesteps).
    for _ in range(4000):
        fa.run_one_step()
        sp.run_one_step(dt=dt)
        z[mg.core_nodes] += U * dt  # m
        br[mg.core_nodes] = z[mg.core_nodes] - soil[mg.core_nodes]

    # compare numerical and analytical slope solutions
    num_slope = mg.at_node["topographic__steepest_slope"][mg.core_nodes]
    analytical_slope = np.power(U / K_br, 1.0 / n_sp) * np.power(
        mg.at_node["drainage_area"][mg.core_nodes], -m_sp / n_sp
    )

    # test for match with analytical slope-area relationship
    testing.assert_array_almost_equal(
        num_slope,
        analytical_slope,
        decimal=8,
        err_msg="SpaceLargeScaleEroder detachment-limited test failed",
        verbose=True,
    )


@pytest.mark.slow
def test_matches_transport_solution():
    # %%
    """
    Test that model matches the transport-limited analytical solution
    for slope/area relationship at steady state: S=((U * v_s) / (K_sed * A^m)
    + U / (K_sed * A^m))^(1/n).

    Also test that model matches the analytical solution for steady-state
    sediment flux: Qs = U * A * (1 - phi).
    """

    # set up a 5x5 grid with one open outlet node and low initial elevations.
    nr = 5
    nc = 5
    mg = RasterModelGrid((nr, nc), xy_spacing=10.0)

    z = mg.add_zeros("topographic__elevation", at="node")
    br = mg.add_zeros("bedrock__elevation", at="node")
    soil = mg.add_zeros("soil__depth", at="node")

    mg["node"]["topographic__elevation"] += (
        mg.node_y / 100000 + mg.node_x / 100000 + np.random.rand(len(mg.node_y)) / 10000
    )
    mg.set_closed_boundaries_at_grid_edges(
        bottom_is_closed=True,
        left_is_closed=True,
        right_is_closed=True,
        top_is_closed=True,
    )
    mg.set_watershed_boundary_condition_outlet_id(
        0, mg["node"]["topographic__elevation"], -9999.0
    )
    soil[:] += 100.0  # initial soil depth of 100 m
    br[:] = z[:]
    z[:] += soil[:]

    # Create a D8 flow handler
    fa = FlowAccumulator(
        mg, flow_director="D8", depression_finder="DepressionFinderAndRouter"
    )

    # Parameter values for detachment-limited test
    K_sed = 0.01
    U = 0.0001
    dt = 1.0
    F_f = 1.0  # all detached rock disappears; detachment-ltd end-member
    m_sp = 0.5
    n_sp = 1.0
    v_s = 0.5
    phi = 0.5

    # Instantiate the SpaceLargeScaleEroder component...
    sp = SpaceLargeScaleEroder(
        mg,
        K_sed=K_sed,
        K_br=0.01,
        F_f=F_f,
        phi=phi,
        H_star=1.0,
        v_s=v_s,
        m_sp=m_sp,
        n_sp=n_sp,
        sp_crit_sed=0,
        sp_crit_br=0,
    )

    # ... and run it to steady state (5000x1-year timesteps).
    for _ in range(5000):
        fa.run_one_step()
        sp.run_one_step(dt=dt)
        br[mg.core_nodes] += U * dt  # m
        soil[
            0
        ] = 100.0  # enforce constant soil depth at boundary to keep lowering steady
        z[:] = br[:] + soil[:]

    # compare numerical and analytical slope solutions
    num_slope = mg.at_node["topographic__steepest_slope"][mg.core_nodes]
    analytical_slope = np.power(
        (
            (U * v_s * (1 - phi))
            / (K_sed * np.power(mg.at_node["drainage_area"][mg.core_nodes], m_sp))
        )
        + (
            (U * (1 - phi))
            / (K_sed * np.power(mg.at_node["drainage_area"][mg.core_nodes], m_sp))
        ),
        1.0 / n_sp,
    )

    # test for match with analytical slope-area relationship
    testing.assert_array_almost_equal(
        num_slope,
        analytical_slope,
        decimal=8,
        err_msg="SpaceLargeScaleEroder transport-limited slope-area test failed",
        verbose=True,
    )

    # compare numerical and analytical sediment flux solutions
    num_sedflux = mg.at_node["sediment__outflux"][mg.core_nodes]
    analytical_sedflux = U * mg.at_node["drainage_area"][mg.core_nodes] * (1 - phi)

    # test for match with anakytical sediment flux
    testing.assert_array_almost_equal(
        num_sedflux,
        analytical_sedflux,
        decimal=8,
        err_msg="SpaceLargeScaleEroder transport-limited sediment flux test failed",
        verbose=True,
    )
    # %%


@pytest.mark.slow
def test_matches_bedrock_alluvial_solution():
    # %%
    """
    Test that model matches the bedrock-alluvial analytical solution
    for slope/area relationship at steady state:
    S=((U * v_s * (1 - F_f)) / (K_sed * A^m) + U / (K_br * A^m))^(1/n).

    Also test that the soil depth everywhere matches the bedrock-alluvial
    analytical solution at steady state:
    H = -H_star * ln(1 - (v_s / (K_sed / (K_br * (1 - F_f)) + v_s))).
    """

    # set up a 5x5 grid with one open outlet node and low initial elevations.
    nr = 5
    nc = 5
    mg = RasterModelGrid((nr, nc), xy_spacing=10.0)

    z = mg.add_zeros("topographic__elevation", at="node")
    br = mg.add_zeros("bedrock__elevation", at="node")
    soil = mg.add_zeros("soil__depth", at="node")

    mg["node"]["topographic__elevation"] += (
        mg.node_y / 100000 + mg.node_x / 100000 + np.random.rand(len(mg.node_y)) / 10000
    )
    mg.set_closed_boundaries_at_grid_edges(
        bottom_is_closed=True,
        left_is_closed=True,
        right_is_closed=True,
        top_is_closed=True,
    )
    mg.set_watershed_boundary_condition_outlet_id(
        0, mg["node"]["topographic__elevation"], -9999.0
    )
    soil[:] += 0.0  # initial condition of no soil depth.
    br[:] = z[:]
    z[:] += soil[:]

    # Create a D8 flow handler
    fa = FlowAccumulator(
        mg, flow_director="D8", depression_finder="DepressionFinderAndRouter"
    )

    # Parameter values for detachment-limited test
    K_br = 0.002
    K_sed = 0.002
    U = 0.0001
    dt = 10.0
    F_f = 0.2  # all detached rock disappears; detachment-ltd end-member
    m_sp = 0.5
    n_sp = 1.0
    v_s = 0.25
    H_star = 0.1

    # Instantiate the SpaceLargeScaleEroder component...
    sp = SpaceLargeScaleEroder(
        mg,
        K_sed=K_sed,
        K_br=K_br,
        F_f=F_f,
        phi=0.0,
        H_star=H_star,
        v_s=v_s,
        m_sp=m_sp,
        n_sp=n_sp,
        sp_crit_sed=0,
        sp_crit_br=0,
    )

    # ... and run it to steady state (10000x1-year timesteps).
    for _ in range(10000):
        fa.run_one_step()
        sp.run_one_step(dt=dt)
        br[mg.core_nodes] += U * dt  # m
        soil[0] = 0.0  # enforce 0 soil depth at boundary to keep lowering steady
        z[:] = br[:] + soil[:]

    # compare numerical and analytical slope solutions
    num_slope = mg.at_node["topographic__steepest_slope"][mg.core_nodes]
    analytical_slope = np.power(
        (
            (U * v_s * (1 - F_f))
            / (K_sed * np.power(mg.at_node["drainage_area"][mg.core_nodes], m_sp))
        )
        + (U / (K_br * np.power(mg.at_node["drainage_area"][mg.core_nodes], m_sp))),
        1.0 / n_sp,
    )

    # test for match with analytical slope-area relationship
    testing.assert_array_almost_equal(
        num_slope,
        analytical_slope,
        decimal=8,
        err_msg="SpaceLargeScaleEroder bedrock-alluvial slope-area test failed",
        verbose=True,
    )

    # compare numerical and analytical soil depth solutions
    num_h = mg.at_node["soil__depth"][mg.core_nodes]
    analytical_h = -H_star * np.log(1 - (v_s / (K_sed / (K_br * (1 - F_f)) + v_s)))

    # test for match with analytical sediment depth
    testing.assert_array_almost_equal(
        num_h,
        analytical_h,
        decimal=5,
        err_msg="SpaceLargeScaleEroder bedrock-alluvial soil thickness test failed",
        verbose=True,
    )
    # %%


def test_can_run_with_hex():
    """Test that model can run with hex model grid."""
    # %%
    # Set up a 5x5 grid with open boundaries and low initial elevations.
    mg = HexModelGrid((7, 7))
    z = mg.add_zeros("topographic__elevation", at="node")
    _ = mg.add_zeros("soil__depth", at="node")
    z[:] = 0.01 * mg.x_of_node

    # Create a D8 flow handler
    fa = FlowAccumulator(mg, flow_director="FlowDirectorSteepest")

    # Parameter values for test 1
    U = 0.001
    dt = 10.0

    # Create the SpaceLargeScaleEroder component...
    sp = SpaceLargeScaleEroder(
        mg,
        K_sed=0.00001,
        K_br=0.00000000001,
        F_f=0.5,
        phi=0.1,
        H_star=1.0,
        v_s=0.001,
        m_sp=0.5,
        n_sp=1.0,
        sp_crit_sed=0,
        sp_crit_br=0,
    )

    # ... and run it to steady state.
    for _ in range(2000):
        fa.run_one_step()
        sp.run_one_step(dt=dt)
        z[mg.core_nodes] += U * dt


@pytest.mark.skipif(not with_richdem, reason="richdem is not installed")
def test_matches_detachment_solution_PF():
    # %%
    """
    Test that model matches the detachment-limited analytical solution
    for slope/area relationship at steady state: S=(U/K_br)^(1/n)*A^(-m/n).
    """

    # set up a 5x5 grid with one open outlet node and low initial elevations.
    nr = 5
    nc = 5
    mg = RasterModelGrid((nr, nc), xy_spacing=10.0)

    z = mg.add_zeros("topographic__elevation", at="node")
    br = mg.add_zeros("bedrock__elevation", at="node")
    soil = mg.add_zeros("soil__depth", at="node")

    mg["node"]["topographic__elevation"] += (
        mg.node_y / 10000 + mg.node_x / 10000 + np.random.rand(len(mg.node_y)) / 10000
    )
    mg.set_closed_boundaries_at_grid_edges(
        bottom_is_closed=True,
        left_is_closed=True,
        right_is_closed=True,
        top_is_closed=True,
    )
    mg.set_watershed_boundary_condition_outlet_id(
        0, mg["node"]["topographic__elevation"], -9999.0
    )
    br[:] = z[:] - soil[:]

    fa = PriorityFloodFlowRouter(
        mg, surface="topographic__elevation", flow_metric="D8", suppress_out=True
    )
    fa.run_one_step()

    # Parameter values for detachment-limited test
    K_br = 0.01
    U = 0.0001
    dt = 1.0
    F_f = 1.0  # all detached rock disappears; detachment-ltd end-member
    m_sp = 0.5
    n_sp = 1.0

    # Instantiate the SpaceLargeScaleEroder component...
    sp = SpaceLargeScaleEroder(
        mg,
        K_sed=0.00001,
        K_br=K_br,
        F_f=F_f,
        phi=0.1,
        H_star=1.0,
        v_s=0.001,
        v_s_lake=0.001,
        m_sp=m_sp,
        n_sp=n_sp,
        sp_crit_sed=0,
        sp_crit_br=0,
    )

    # ... and run it to steady state (2000x1-year timesteps).
    for _ in range(2000):
        fa.run_one_step()
        sp.run_one_step(dt=dt)
        z[mg.core_nodes] += U * dt  # m
        br[mg.core_nodes] = z[mg.core_nodes] - soil[mg.core_nodes]

    # compare numerical and analytical slope solutions
    num_slope = mg.at_node["topographic__steepest_slope"][mg.core_nodes]
    analytical_slope = np.power(U / K_br, 1.0 / n_sp) * np.power(
        mg.at_node["drainage_area"][mg.core_nodes], -m_sp / n_sp
    )

    # test for match with analytical slope-area relationship
    testing.assert_array_almost_equal(
        num_slope,
        analytical_slope,
        decimal=8,
        err_msg="SpaceLargeScaleEroder detachment-limited test failed",
        verbose=True,
    )


# %%
@pytest.mark.slow
@pytest.mark.skipif(not with_richdem, reason="richdem is not installed")
def test_matches_transport_solution_PF():
    """
    Test that model matches the transport-limited analytical solution
    for slope/area relationship at steady state: S=((U * v_s) / (K_sed * A^m)
    + U / (K_sed * A^m))^(1/n).

    Also test that model matches the analytical solution for steady-state
    sediment flux: Qs = U * A * (1 - phi).
    """
    # %%
    # set up a 5x5 grid with one open outlet node and low initial elevations.
    nr = 5
    nc = 5
    mg = RasterModelGrid((nr, nc), xy_spacing=10.0)

    z = mg.add_zeros("topographic__elevation", at="node")
    br = mg.add_zeros("bedrock__elevation", at="node")
    soil = mg.add_zeros("soil__depth", at="node")

    mg["node"]["topographic__elevation"] += (
        mg.node_y / 100000 + mg.node_x / 100000 + np.random.rand(len(mg.node_y)) / 10000
    )
    mg.set_closed_boundaries_at_grid_edges(
        bottom_is_closed=True,
        left_is_closed=True,
        right_is_closed=True,
        top_is_closed=True,
    )
    mg.set_watershed_boundary_condition_outlet_id(
        0, mg["node"]["topographic__elevation"], -9999.0
    )
    soil[:] += 100.0  # initial soil depth of 100 m
    br[:] = z[:]
    z[:] += soil[:]

    # Create a D8 flow handler
    fa = PriorityFloodFlowRouter(
        mg, surface="topographic__elevation", flow_metric="D8", suppress_out=True
    )
    fa.run_one_step()

    # Parameter values for detachment-limited test
    K_sed = 0.01
    U = 0.0001
    dt = 1.0
    F_f = 1.0  # all detached rock disappears; detachment-ltd end-member
    m_sp = 0.5
    n_sp = 1.0
    v_s = 0.5
    phi = 0.5

    # Instantiate the SpaceLargeScaleEroder component...
    sp = SpaceLargeScaleEroder(
        mg,
        K_sed=K_sed,
        K_br=0.01,
        F_f=F_f,
        phi=phi,
        H_star=1.0,
        v_s=v_s,
        m_sp=m_sp,
        n_sp=n_sp,
        sp_crit_sed=0,
        sp_crit_br=0,
    )

    # ... and run it to steady state (5000x1-year timesteps).
    for _ in range(5000):
        fa.run_one_step()
        sp.run_one_step(dt=dt)
        br[mg.core_nodes] += U * dt  # m
        soil[
            0
        ] = 100.0  # enforce constant soil depth at boundary to keep lowering steady
        z[:] = br[:] + soil[:]

    # compare numerical and analytical slope solutions
    num_slope = mg.at_node["topographic__steepest_slope"][mg.core_nodes]
    analytical_slope = np.power(
        (
            (U * v_s * (1 - phi))
            / (K_sed * np.power(mg.at_node["drainage_area"][mg.core_nodes], m_sp))
        )
        + (
            (U * (1 - phi))
            / (K_sed * np.power(mg.at_node["drainage_area"][mg.core_nodes], m_sp))
        ),
        1.0 / n_sp,
    )

    # test for match with analytical slope-area relationship
    testing.assert_array_almost_equal(
        num_slope,
        analytical_slope,
        decimal=8,
        err_msg="SpaceLargeScaleEroder transport-limited slope-area test failed",
        verbose=True,
    )

    # compare numerical and analytical sediment flux solutions
    num_sedflux = mg.at_node["sediment__outflux"][mg.core_nodes]
    analytical_sedflux = U * mg.at_node["drainage_area"][mg.core_nodes] * (1 - phi)

    # test for match with anakytical sediment flux
    testing.assert_array_almost_equal(
        num_sedflux,
        analytical_sedflux,
        decimal=8,
        err_msg="SpaceLargeScaleEroder transport-limited sediment flux test failed",
        verbose=True,
    )


# %%
@pytest.mark.slow
@pytest.mark.skipif(not with_richdem, reason="richdem is not installed")
def test_matches_bedrock_alluvial_solution_PF():
    """
    Test that model matches the bedrock-alluvial analytical solution
    for slope/area relationship at steady state:
    S=((U * v_s * (1 - F_f)) / (K_sed * A^m) + U / (K_br * A^m))^(1/n).

    Also test that the soil depth everywhere matches the bedrock-alluvial
    analytical solution at steady state:
    H = -H_star * ln(1 - (v_s / (K_sed / (K_br * (1 - F_f)) + v_s))).
    """
    # %%
    # set up a 5x5 grid with one open outlet node and low initial elevations.
    nr = 5
    nc = 5
    mg = RasterModelGrid((nr, nc), xy_spacing=10.0)

    z = mg.add_zeros("topographic__elevation", at="node")
    br = mg.add_zeros("bedrock__elevation", at="node")
    soil = mg.add_zeros("soil__depth", at="node")

    mg["node"]["topographic__elevation"] += (
        mg.node_y / 100000 + mg.node_x / 100000 + np.random.rand(len(mg.node_y)) / 10000
    )
    mg.set_closed_boundaries_at_grid_edges(
        bottom_is_closed=True,
        left_is_closed=True,
        right_is_closed=True,
        top_is_closed=True,
    )
    mg.set_watershed_boundary_condition_outlet_id(
        0, mg["node"]["topographic__elevation"], -9999.0
    )
    soil[:] += 0.0  # initial condition of no soil depth.
    br[:] = z[:]
    z[:] += soil[:]

    # Create a D8 flow handler
    fa = PriorityFloodFlowRouter(
        mg, surface="topographic__elevation", flow_metric="D8", suppress_out=True
    )
    fa.run_one_step()

    # Parameter values for detachment-limited test
    K_br = 0.002
    K_sed = 0.002
    U = 0.0001
    dt = 10.0
    F_f = 0.2  # all detached rock disappears; detachment-ltd end-member
    m_sp = 0.5
    n_sp = 1.0
    v_s = 0.25
    H_star = 0.1

    # Instantiate the SpaceLargeScaleEroder component...
    sp = SpaceLargeScaleEroder(
        mg,
        K_sed=K_sed,
        K_br=K_br,
        F_f=F_f,
        phi=0.0,
        H_star=H_star,
        v_s=v_s,
        m_sp=m_sp,
        n_sp=n_sp,
        sp_crit_sed=0,
        sp_crit_br=0,
    )

    # ... and run it to steady state (10000x1-year timesteps).
    for _ in range(10000):
        fa.run_one_step()
        sp.run_one_step(dt=dt)
        br[mg.core_nodes] += U * dt  # m
        soil[0] = 0.0  # enforce 0 soil depth at boundary to keep lowering steady
        z[:] = br[:] + soil[:]

    # compare numerical and analytical slope solutions
    num_slope = mg.at_node["topographic__steepest_slope"][mg.core_nodes]
    analytical_slope = np.power(
        (
            (U * v_s * (1 - F_f))
            / (K_sed * np.power(mg.at_node["drainage_area"][mg.core_nodes], m_sp))
        )
        + (U / (K_br * np.power(mg.at_node["drainage_area"][mg.core_nodes], m_sp))),
        1.0 / n_sp,
    )

    # test for match with analytical slope-area relationship
    testing.assert_array_almost_equal(
        num_slope,
        analytical_slope,
        decimal=8,
        err_msg="SpaceLargeScaleEroder bedrock-alluvial slope-area test failed",
        verbose=True,
    )

    # compare numerical and analytical soil depth solutions
    num_h = mg.at_node["soil__depth"][mg.core_nodes]
    analytical_h = -H_star * np.log(1 - (v_s / (K_sed / (K_br * (1 - F_f)) + v_s)))

    # test for match with analytical sediment depth
    testing.assert_array_almost_equal(
        num_h,
        analytical_h,
        decimal=5,
        err_msg="SpaceLargeScaleEroder bedrock-alluvial soil thickness test failed",
        verbose=True,
    )
    # %%


@pytest.mark.slow
def test_MassBalance():
    # %%
    # set up a 15x15 grid with one open outlet node and low initial elevations.
    nr = 15
    nc = 15
    mg = RasterModelGrid((nr, nc), xy_spacing=10.0)

    z = mg.add_zeros("topographic__elevation", at="node")
    br = mg.add_zeros("bedrock__elevation", at="node")
    soil = mg.add_zeros("soil__depth", at="node")

    mg["node"]["topographic__elevation"] += (
        mg.node_y / 100000 + mg.node_x / 100000 + np.random.rand(len(mg.node_y)) / 10000
    )
    mg.set_closed_boundaries_at_grid_edges(
        bottom_is_closed=True,
        left_is_closed=True,
        right_is_closed=True,
        top_is_closed=True,
    )
    mg.set_watershed_boundary_condition_outlet_id(
        0, mg["node"]["topographic__elevation"], -9999.0
    )
    soil[:] += 0.0  # initial condition of no soil depth.
    br[:] = z[:]
    z[:] += soil[:]

    # Create a D8 flow handler
    # fa = PriorityFloodFlowRouter(
    #   mg, surface="topographic__elevation", flow_metric = 'D8',suppress_out=True
    # )
    # fa.run_one_step()

    # Create a D8 flow handler
    fa = FlowAccumulator(
        mg, flow_director="D8", depression_finder="DepressionFinderAndRouter"
    )

    # Parameter values for detachment-limited test
    K_br = 0.002
    K_sed = 0.002
    U = 0.0001
    dt = 10.0
    F_f = 0.2  # all detached rock disappears; detachment-ltd end-member
    m_sp = 0.5
    n_sp = 1.0
    v_s = 0.25
    H_star = 0.1

    # Instantiate the Space component...
    sp = SpaceLargeScaleEroder(
        mg,
        K_sed=K_sed,
        K_br=K_br,
        F_f=F_f,
        phi=0.0,
        H_star=H_star,
        v_s=v_s,
        m_sp=m_sp,
        n_sp=n_sp,
        sp_crit_sed=0,
        sp_crit_br=0,
    )
    # Get values before run
    z = mg.at_node["topographic__elevation"]
    br = mg.at_node["bedrock__elevation"]
    H = mg.at_node["soil__depth"]
    cores = mg.core_nodes
    area = mg.cell_area_at_node
    # ... and run it to steady state (10000x1-year timesteps).
    for _ in range(10000):
        fa.run_one_step()
        soil_B = cp.deepcopy(H)
        bed_B = cp.deepcopy(br)
        vol_SSY_riv, V_leaving_riv = sp.run_one_step(dt=dt)
        diff_MB = (
            np.sum((bed_B[cores] - br[cores]) * area[cores])
            + np.sum((soil_B[cores] - H[cores]) * area[cores]) * (1 - sp._phi)
            - vol_SSY_riv * dt
            - V_leaving_riv
        )

        br[mg.core_nodes] += U * dt  # m
        soil[0] = 0.0  # enforce 0 soil depth at boundary to keep lowering steady
        z[:] = br[:] + soil[:]

        # Test Every iteration
        testing.assert_array_almost_equal(
            z[cores],
            br[cores] + H[cores],
            decimal=5,
            err_msg="Topography does not equal sum of bedrock and soil! Decrease timestep",
            verbose=True,
        )
        testing.assert_array_less(
            abs(diff_MB),
            1e-8 * mg.number_of_nodes,
            err_msg=(
                "Mass balance error SpaceLargeScaleEroder! Try to resolve by "
                "becreasing timestep"
            ),
            verbose=True,
        )



"""
Unit tests for landlab.data_record.data_record.DataRecord
Dimension = time and item_id

Last updated 8/24/2018

"""

import numpy as np

from landlab import RasterModelGrid

grid = RasterModelGrid((3, 3))
shape = (3, 3)
time = [0.0]
my_items3 = {
    "grid_element": np.array([["node"], ["link"]]),
    "element_id": np.array([[1], [3]]),
}
my_data_vars = {
    "mean_elevation": (["time"], [110.0]),
    "item_size": (["item_id", "time"], np.array([[0.3], [0.4]])),
}
#    return DataRecord(grid, time=time, items=my_items3, data_vars=my_data_vars )


def test_dr_name(dr_2dim):
    assert dr_2dim._name == "DataRecord"


def test_grid_shape(dr_2dim):
    assert dr_2dim._grid.number_of_node_rows == shape[0]
    assert dr_2dim._grid.number_of_node_columns == shape[1]


def test_permitted_locations(dr_2dim):
    assert dr_2dim._permitted_locations == grid.groups


def test_coordinates(dr_2dim):
    assert len(dr_2dim.dataset.dims) == 2
    assert list(dr_2dim.dataset.time.values) == time
    assert list(dr_2dim.time_coordinates) == time
    assert list(dr_2dim.dataset.item_id.values) == [0, 1]
    assert list(dr_2dim.item_coordinates) == [0, 1]
    # properties:
    assert dr_2dim.number_of_timesteps == 1
    assert dr_2dim.number_of_items == 2
    assert dr_2dim.earliest_time == 0.0
    assert dr_2dim.latest_time == 0.0
    assert np.isnan(dr_2dim.prior_time)
    dr_2dim.add_record(time=[10.0, 20.0], new_record={"new_data": ((), (120.0, 130.0))})
    assert dr_2dim.prior_time == 10.0


def test_variable_names(dr_2dim):
    assert sorted(dr_2dim.variable_names) == sorted(
        ["element_id", "grid_element", "item_size", "mean_elevation"]
    )


def test_add_record(dr_2dim):
    dr_2dim.add_record(
        time=[10.0],
        item_id=[1],
        new_item_loc={
            "grid_element": np.array([["cell"]]),
            "element_id": np.array([[0]]),
        },
    )
    dr_2dim.add_record(
        time=[20.0], new_record={"mean_elevation": (["time"], np.array([130.0]))}
    )
    assert (
        dr_2dim.dataset["grid_element"].values[1, 1],
        dr_2dim.dataset["mean_elevation"].values[2],
    ) == ("cell", 130.0)
    assert np.isnan(dr_2dim.dataset["element_id"].values[1, 2])
    dr_2dim.add_record(
        time=[10.0],
        item_id=[1],
        new_record={"size": (["item_id", "time"], np.array([[0.3]]))},
    )
    assert np.isnan(dr_2dim.dataset["size"].values[0, 0])


def test_add_item(dr_2dim):
    dr_2dim.add_item(
        time=[10.0],
        new_item={
            "grid_element": np.array([["node"], ["cell"]]),
            "element_id": np.array([[2], [0]]),
        },
        new_item_spec={"size": (["item_id"], [10, 5])},
    )
    assert (
        dr_2dim.dataset["grid_element"].values[3, 1],
        dr_2dim.dataset["element_id"].values[2, 1],
        dr_2dim.dataset["size"].values[3],
    ) == ("cell", 2.0, 5.0)


def test_get_data(dr_2dim):
    assert (
        dr_2dim.get_data(time=[0.0], item_id=[1], data_variable="grid_element")
        == "link"
    )
    assert dr_2dim.get_data(data_variable="mean_elevation") == [110.0]


def test_set_data(dr_2dim):
    dr_2dim.set_data(
        time=[0.0], item_id=[1], data_variable="grid_element", new_value="node"
    )
    dr_2dim.set_data(time=[0.0], data_variable="mean_elevation", new_value=150.0)
    assert all(dr_2dim.dataset["grid_element"].values == "node")
    assert dr_2dim.dataset["mean_elevation"].values[0] == 150.0


def test_ffill_grid_element_and_id(dr_2dim):
    dr_2dim.add_record(
        time=[20.0], new_record={"mean_elevation": (["time"], np.array([130.0]))}
    )
    dr_2dim.ffill_grid_element_and_id()
    assert dr_2dim.dataset["grid_element"].values[0, 0] == (
        dr_2dim.dataset["grid_element"].values[0, 1]
    )
    assert dr_2dim.dataset["element_id"].values[0, 0] == (
        dr_2dim.dataset["element_id"].values[0, 1]
    )
